"""Tests for _build_query_params() — $filter vs $search logic."""

from __future__ import annotations

from outlook_cli.client import _build_query_params


class TestNoFilters:
    def test_returns_empty(self):
        f, s, needs = _build_query_params()
        assert f == ""
        assert s == ""
        assert needs is False


class TestFilterPath:
    """Pure $filter — supports $orderby."""

    def test_unread_only(self):
        f, s, needs = _build_query_params(unread_only=True)
        assert "IsRead eq false" in f
        assert needs is False

    def test_after_date(self):
        f, s, needs = _build_query_params(filter_after="2026-03-01")
        assert "ReceivedDateTime ge 2026-03-01T00:00:00Z" in f
        assert needs is False

    def test_before_date(self):
        f, s, needs = _build_query_params(filter_before="2026-03-08")
        assert "ReceivedDateTime lt 2026-03-08T23:59:59Z" in f

    def test_category_filter(self):
        f, s, needs = _build_query_params(filter_category="Finance")
        assert "Categories/any(c:c eq 'Finance')" in f
        assert needs is False

    def test_combined_filter(self):
        f, s, needs = _build_query_params(unread_only=True, filter_after="2026-03-01")
        assert "IsRead eq false" in f
        assert "ReceivedDateTime ge" in f
        assert " and " in f
        assert needs is False


class TestSearchPath:
    """KQL $search — can't use $orderby."""

    def test_from_filter(self):
        f, s, needs = _build_query_params(filter_from="alice")
        assert "from:alice" in s
        assert needs is True
        assert f == ""

    def test_subject_filter(self):
        f, s, needs = _build_query_params(filter_subject="Q4 Report")
        assert "subject:Q4 Report" in s
        assert needs is True

    def test_has_attachments(self):
        f, s, needs = _build_query_params(filter_has_attachments=True)
        assert "hasattachments:true" in s
        assert needs is True

    def test_text_plus_date_uses_search(self):
        """When text filters exist, dates go into KQL $search too."""
        f, s, needs = _build_query_params(
            filter_from="bob",
            filter_after="2026-03-01",
        )
        assert needs is True
        assert "from:bob" in s
        assert "received>=2026-03-01" in s
        assert f == ""

    def test_text_plus_category_uses_search(self):
        f, s, needs = _build_query_params(
            filter_subject="invoice",
            filter_category="Finance",
        )
        assert needs is True
        assert "subject:invoice" in s
        assert 'category:"Finance"' in s

    def test_text_plus_unread_uses_search(self):
        f, s, needs = _build_query_params(
            filter_from="alice",
            unread_only=True,
        )
        assert needs is True
        assert "isread:false" in s
        assert "from:alice" in s



from __future__ import annotations

from datetime import datetime, timezone

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from .models import Attachment, Attendee, Contact, Email, Event, Folder

console = Console(stderr=True)


def print_inbox(messages: list[Email]) -> None:
    show_categories = any(msg.categories for msg in messages)

    table = Table(show_header=True, header_style="bold cyan", box=None, pad_edge=False)
    table.add_column("#", style="dim", width=5, justify="right")
    table.add_column("From", width=28, no_wrap=True)
    table.add_column("Subject", ratio=1, no_wrap=True, overflow="ellipsis")
    if show_categories:
        table.add_column("Category", width=16, no_wrap=True, overflow="ellipsis")
    table.add_column("Date", width=9, no_wrap=True, justify="right")
    table.add_column("", width=2)  # flags

    for msg in messages:
        flags = ""
        if not msg.is_read:
            flags += "*"
        if msg.has_attachments:
            flags += "@"

        style = "bold" if not msg.is_read else ""
        row = [
            str(msg.display_num),
            _truncate(str(msg.sender), 25),
            _truncate(msg.subject, 50),
        ]
        if show_categories:
            row.append(_truncate(", ".join(msg.categories), 16) if msg.categories else "")
        row.extend([_format_date(msg.received), flags])
        table.add_row(*row, style=style)

    console.print(table)


def print_email(email: Email) -> None:
    header = (
        f"[bold]From:[/bold] {email.sender}\n"
        f"[bold]To:[/bold] {', '.join(str(r) for r in email.to)}\n"
    )
    if email.cc:
        header += f"[bold]Cc:[/bold] {', '.join(str(r) for r in email.cc)}\n"
    header += f"[bold]Date:[/bold] {email.received.strftime('%Y-%m-%d %H:%M')}\n"
    if email.categories:
        header += f"[bold]Categories:[/bold] {', '.join(email.categories)}\n"
    header += f"[bold]Subject:[/bold] {email.subject}"

    body = _html_to_text(email.body) if email.body_type == "HTML" else email.body

    console.print(Panel(header, title=f"Message #{email.display_num}", border_style="cyan"))
    console.print()
    console.print(body)


def print_thread(messages: list[Email]) -> None:
    """Print a conversation thread — all messages in chronological order."""
    console.print(f"[bold cyan]Thread ({len(messages)} messages)[/bold cyan]")
    console.print()
    for i, email in enumerate(messages):
        is_last = i == len(messages) - 1
        sender = str(email.sender)
        date = email.received.strftime("%Y-%m-%d %H:%M")
        read_marker = "" if email.is_read else " [bold yellow]*[/bold yellow]"

        header = f"[bold]#{email.display_num}[/bold]  [dim]{date}[/dim]  {sender}{read_marker}"
        console.print(header)

        body = _html_to_text(email.body) if email.body_type == "HTML" else email.body
        body = body.strip()
        if body:
            # Indent body and truncate long messages
            lines = body.split("\n")
            if len(lines) > 20:
                lines = lines[:20] + [f"  [dim]... ({len(lines) - 20} more lines)[/dim]"]
            for line in lines:
                console.print(f"  {line}")

        if not is_last:
            console.print(f"  [dim]{'─' * 60}[/dim]")
            console.print()


def print_email_raw(email: Email) -> None:
    """Print raw HTML body."""
    console.print(email.body)


def print_folders(folders: list[Folder]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None)
    table.add_column("Name", min_width=20)
    table.add_column("Unread", justify="right", width=8)
    table.add_column("Total", justify="right", width=8)

    for f in folders:
        style = "bold" if f.unread_count > 0 else ""
        table.add_row(f.name, str(f.unread_count), str(f.total_count), style=style)

    console.print(table)


def print_attachments(attachments: list[Attachment]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None)
    table.add_column("#", width=4, justify="right")
    table.add_column("Name", min_width=30)
    table.add_column("Type", width=20)
    table.add_column("Size", width=10, justify="right")

    for i, att in enumerate(attachments, 1):
        table.add_row(str(i), att.name, att.content_type, _format_size(att.size))

    console.print(table)


def print_events(events: list[Event]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None, pad_edge=False)
    table.add_column("#", style="dim", width=5, justify="right")
    table.add_column("Date", width=12)
    table.add_column("Time", width=13)
    table.add_column("Subject", ratio=1, no_wrap=True, overflow="ellipsis")
    table.add_column("Location", max_width=20, no_wrap=True)
    table.add_column("Ppl", width=4, justify="right")

    for ev in events:
        if ev.is_all_day:
            time_str = "All day"
        else:
            time_str = f"{ev.start.strftime('%H:%M')}-{ev.end.strftime('%H:%M')}"
        ppl = str(len(ev.attendees)) if ev.attendees else ""
        table.add_row(
            str(ev.display_num) if ev.display_num else "",
            ev.start.strftime("%Y-%m-%d"),
            time_str,
            _truncate(ev.subject, 45),
            _truncate(ev.location, 20),
            ppl,
        )

    console.print(table)


def print_event_detail(event: Event) -> None:
    header = f"[bold]Subject:[/bold] {event.subject}\n"
    if event.is_all_day:
        header += f"[bold]When:[/bold] {event.start.strftime('%Y-%m-%d')} (All day)\n"
    else:
        header += f"[bold]Start:[/bold] {event.start.strftime('%Y-%m-%d %H:%M')}\n"
        header += f"[bold]End:[/bold] {event.end.strftime('%Y-%m-%d %H:%M')}\n"
    if event.location:
        header += f"[bold]Location:[/bold] {event.location}\n"
    header += f"[bold]Organizer:[/bold] {event.organizer}\n"
    header += f"[bold]Show as:[/bold] {event.show_as}\n"
    if event.is_online_meeting and event.online_meeting_url:
        header += f"[bold]Online:[/bold] {event.online_meeting_url}\n"
    if event.categories:
        header += f"[bold]Categories:[/bold] {', '.join(event.categories)}\n"
    if event.response_status:
        header += f"[bold]Your response:[/bold] {event.response_status}\n"
    if event.recurrence:
        header += f"[bold]Recurrence:[/bold] {_format_recurrence(event.recurrence)}\n"
    if event.event_type and event.event_type != "SingleInstance":
        header += f"[bold]Type:[/bold] {event.event_type}\n"
    if event.is_cancelled:
        header += "[bold red]CANCELLED[/bold red]\n"

    num = f"Event #{event.display_num}" if event.display_num else "Event"
    console.print(Panel(header.rstrip(), title=num, border_style="cyan"))

    if event.attendees:
        console.print(f"\n[bold]Attendees ({len(event.attendees)}):[/bold]")
        for att in event.attendees:
            resp_icon = {"Accepted": "[green]v[/green]", "Declined": "[red]x[/red]",
                         "TentativelyAccepted": "[yellow]?[/yellow]"}.get(att.response, "[dim]-[/dim]")
            att_type = f" [dim]({att.type})[/dim]" if att.type != "Required" else ""
            console.print(f"  {resp_icon} {att.email}{att_type}")

    if event.body_preview:
        console.print(f"\n{event.body_preview}")


def print_calendars(calendars: list[dict]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None)
    table.add_column("Name", min_width=25)
    table.add_column("Owner", min_width=30)
    table.add_column("Color", width=12)
    table.add_column("Edit", width=5, justify="center")

    for cal in calendars:
        owner = cal.get("Owner", {}).get("Address", "")
        can_edit = "Yes" if cal.get("CanEdit") else "No"
        table.add_row(cal.get("Name", ""), owner, cal.get("Color", ""), can_edit)

    console.print(table)


def print_meeting_suggestions(suggestions: list[dict]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None)
    table.add_column("#", width=4, justify="right")
    table.add_column("Start", width=18)
    table.add_column("End", width=18)
    table.add_column("Confidence", width=10, justify="right")
    table.add_column("Availability", ratio=1)

    for i, s in enumerate(suggestions, 1):
        slot = s.get("MeetingTimeSlot", {})
        start = slot.get("Start", {}).get("DateTime", "")[:16]
        end = slot.get("End", {}).get("DateTime", "")[:16]
        confidence = f"{s.get('Confidence', 0)}%"
        avail_parts = []
        for att in s.get("AttendeeAvailability", []):
            email = att.get("Attendee", {}).get("EmailAddress", {}).get("Address", "")
            avail = att.get("Availability", "?")
            avail_parts.append(f"{email}={avail}")
        table.add_row(str(i), start, end, confidence, "; ".join(avail_parts))

    console.print(table)


def print_people(people: list[dict]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None)
    table.add_column("Name", min_width=25)
    table.add_column("Email", min_width=30)
    table.add_column("Title", max_width=25)

    for p in people:
        emails = p.get("ScoredEmailAddresses", [])
        email = emails[0].get("Address", "") if emails else ""
        table.add_row(
            p.get("DisplayName", ""),
            email,
            p.get("JobTitle", "") or "",
        )

    console.print(table)


def print_contacts(contacts: list[Contact]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None)
    table.add_column("Name", min_width=25)
    table.add_column("Email", min_width=30)
    table.add_column("Company", max_width=20)
    table.add_column("Title", max_width=20)

    for c in contacts:
        email = c.email_addresses[0].address if c.email_addresses else ""
        table.add_row(c.display_name, email, c.company, c.job_title)

    console.print(table)


def print_categories(categories: list[dict]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None)
    table.add_column("Category", min_width=25)
    table.add_column("Unread", justify="right", width=8)
    table.add_column("Total", justify="right", width=8)

    for c in categories:
        style = "bold" if c.get("UnreadCount", 0) > 0 else ""
        table.add_row(
            c["Category"],
            str(c.get("UnreadCount", 0)),
            str(c.get("ItemCount", 0)),
            style=style,
        )

    console.print(table)


def print_whoami(data: dict) -> None:
    console.print(f"[bold]Name:[/bold]  {data.get('DisplayName', 'N/A')}")
    console.print(f"[bold]Email:[/bold] {data.get('EmailAddress', 'N/A')}")
    console.print(f"[bold]Alias:[/bold] {data.get('Alias', 'N/A')}")


def print_success(msg: str) -> None:
    console.print(f"[green]{msg}[/green]")


def print_error(msg: str) -> None:
    console.print(f"[red]{msg}[/red]")


# ------------------------------------------------------------------
# Helpers
# ------------------------------------------------------------------

def _truncate(s: str, max_len: int) -> str:
    if len(s) <= max_len:
        return s
    return s[: max_len - 1] + "\u2026"


def _format_date(dt: datetime) -> str:
    # Compare in local timezone so "today" matches user's perception
    now_local = datetime.now().astimezone()
    if dt.tzinfo is None:
        from datetime import timezone as tz
        dt = dt.replace(tzinfo=tz.utc)
    dt_local = dt.astimezone(now_local.tzinfo)
    diff = now_local - dt_local
    if dt_local.date() == now_local.date():
        return dt_local.strftime("%H:%M")
    if (now_local.date() - dt_local.date()).days == 1:
        return "Yday"
    if diff.days < 7:
        return dt_local.strftime("%a")  # Mon, Tue...
    if dt_local.year == now_local.year:
        return dt_local.strftime("%d %b")  # 03 Mar
    return dt_local.strftime("%d %b %y")  # 03 Mar 25


def _format_size(size_bytes: int) -> str:
    for unit in ("B", "KB", "MB", "GB"):
        if size_bytes < 1024:
            return f"{size_bytes:.0f}{unit}"
        size_bytes /= 1024
    return f"{size_bytes:.1f}TB"


def _format_recurrence(rec: dict) -> str:
    """Format recurrence dict into human-readable string."""
    pat = rec.get("Pattern", {})
    rng = rec.get("Range", {})
    ptype = pat.get("Type", "")
    interval = pat.get("Interval", 1)

    if ptype == "Daily":
        desc = f"Every {interval} day(s)" if interval > 1 else "Daily"
    elif ptype == "Weekly":
        days = ", ".join(pat.get("DaysOfWeek", []))
        desc = f"Every {interval} week(s) on {days}" if interval > 1 else f"Weekly on {days}"
    elif ptype == "AbsoluteMonthly":
        day = pat.get("DayOfMonth", "?")
        desc = f"Every {interval} month(s) on day {day}" if interval > 1 else f"Monthly on day {day}"
    elif ptype == "RelativeMonthly":
        idx = pat.get("Index", "")
        days = ", ".join(pat.get("DaysOfWeek", []))
        desc = f"Monthly on {idx} {days}"
    elif ptype == "AbsoluteYearly":
        month = pat.get("Month", "?")
        day = pat.get("DayOfMonth", "?")
        desc = f"Yearly on {month}/{day}"
    else:
        desc = ptype

    rtype = rng.get("Type", "")
    if rtype == "Numbered":
        desc += f" ({rng.get('NumberOfOccurrences', '?')} times)"
    elif rtype == "EndDate":
        desc += f" (until {rng.get('EndDate', '?')})"
    elif rtype == "NoEnd":
        desc += " (no end)"

    return desc


def _html_to_text(html: str) -> str:
    """Convert HTML email body to readable text."""
    try:
        from bs4 import BeautifulSoup
        soup = BeautifulSoup(html, "html.parser")
        # Remove style and script tags
        for tag in soup(["style", "script"]):
            tag.decompose()
        return soup.get_text(separator="\n", strip=True)
    except ImportError:
        # Fallback: strip tags with regex
        import re
        return re.sub(r"<[^>]+>", "", html)



"""Folder commands: folders, folder."""

from __future__ import annotations

import click

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    cfg,
    console,
    print_folders,
    print_inbox,
    print_success,
    save_json,
    to_json_envelope,
)


@click.command()
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--output", "-o", type=click.Path(), help="Save output to file")
@_handle_api_error
def folders(as_json: bool, output: str | None):
    """List mail folders."""
    client = _get_client()
    folder_list = client.get_folders()

    if _wants_json(as_json):
        if output:
            save_json(folder_list, output)
            print_success(f"Saved to {output}")
        else:
            click.echo(to_json_envelope(folder_list))
    else:
        print_folders(folder_list)


@click.command()
@click.argument("name")
@click.option("--max", "-n", "max_count", default=None, type=int, help="Number of messages")
@click.option("--unread", is_flag=True, help="Show only unread messages")
@click.option("--from", "from_filter", default=None, help="Filter by sender")
@click.option("--subject", default=None, help="Filter by subject")
@click.option("--after", default=None, help="After date (YYYY-MM-DD)")
@click.option("--before", default=None, help="Before date (YYYY-MM-DD)")
@click.option("--has-attachments", is_flag=True, help="Only messages with attachments")
@click.option("--category", default=None, help="Filter by category name")
@click.option("--no-category", "no_category", is_flag=True, help="Only uncategorized messages")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def folder(
    name: str,
    max_count: int | None,
    unread: bool,
    from_filter: str | None,
    subject: str | None,
    after: str | None,
    before: str | None,
    has_attachments: bool,
    category: str | None,
    no_category: bool,
    as_json: bool,
):
    """Show messages in a specific folder."""
    client = _get_client()
    top = max_count or cfg["max_messages"]
    messages = client.get_messages(
        folder=name,
        top=top,
        unread_only=unread,
        filter_from=from_filter,
        filter_subject=subject,
        filter_after=after,
        filter_before=before,
        filter_has_attachments=has_attachments,
        filter_category=category,
        filter_no_category=no_category,
    )

    if _wants_json(as_json):
        click.echo(to_json_envelope(messages))
    else:
        if not messages:
            print_success(f"No messages found in '{name}'.")
        else:
            console.print(f"[bold cyan]Folder: {name}[/bold cyan]")
            print_inbox(messages)



"""Search commands: search."""

from __future__ import annotations

import click

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    print_error,
    print_inbox,
    save_json,
    to_json_envelope,
    print_success,
)


@click.command()
@click.argument("query")
@click.option("--max", "-n", "max_count", default=25, type=int, help="Max results")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--output", "-o", type=click.Path(), help="Save output to file")
@_handle_api_error
def search(query: str, max_count: int, as_json: bool, output: str | None):
    """Search messages."""
    client = _get_client()
    messages = client.search_messages(query, top=max_count)

    if _wants_json(as_json):
        if output:
            save_json(messages, output)
            print_success(f"Saved to {output}")
        else:
            click.echo(to_json_envelope(messages))
    else:
        if not messages:
            print_error("No results found.")
        else:
            print_inbox(messages)



"""Tests for exceptions.py — hierarchy and error code mapping."""

from __future__ import annotations

from outlook_cli.exceptions import (
    AuthRequiredError,
    OutlookCliError,
    RateLimitError,
    ResourceNotFoundError,
    TokenExpiredError,
    error_code_for_exception,
)


class TestExceptionHierarchy:
    def test_all_inherit_from_base(self):
        for cls in [TokenExpiredError, RateLimitError, ResourceNotFoundError, AuthRequiredError]:
            assert issubclass(cls, OutlookCliError)
            assert issubclass(cls, Exception)

    def test_base_catches_all(self):
        for cls in [TokenExpiredError, RateLimitError, ResourceNotFoundError, AuthRequiredError]:
            try:
                raise cls("test")
            except OutlookCliError:
                pass  # should be caught

    def test_siblings_dont_catch_each_other(self):
        try:
            raise TokenExpiredError("test")
        except ResourceNotFoundError:
            assert False, "TokenExpiredError should NOT be caught by ResourceNotFoundError"
        except TokenExpiredError:
            pass


class TestErrorCodeMapping:
    def test_known_exceptions(self):
        assert error_code_for_exception(TokenExpiredError("x")) == "session_expired"
        assert error_code_for_exception(RateLimitError("x")) == "rate_limited"
        assert error_code_for_exception(ResourceNotFoundError("x")) == "not_found"
        assert error_code_for_exception(AuthRequiredError("x")) == "not_authenticated"

    def test_unknown_exception(self):
        assert error_code_for_exception(ValueError("x")) == "unknown_error"
        assert error_code_for_exception(RuntimeError("x")) == "unknown_error"



import os
from pathlib import Path

# API
BASE_URL = "https://outlook.office.com/api/v2.0/me"
OWA_URL = "https://outlook.office365.com/mail/"
OWA_SERVICE_URL = "https://outlook.office365.com/owa/service.svc"

# User-Agent mimics Outlook Web App
USER_AGENT = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/120.0.0.0 Safari/537.36"
)

# Cache paths
CACHE_DIR = Path(os.environ.get("OUTLOOK_CLI_CACHE", Path.home() / ".cache" / "outlook-cli"))
TOKEN_FILE = CACHE_DIR / "token.json"
BROWSER_STATE_FILE = CACHE_DIR / "browser-state.json"
ID_MAP_FILE = CACHE_DIR / "id_map.json"

# Config paths
CONFIG_DIR = Path(os.environ.get("OUTLOOK_CLI_CONFIG", Path.home() / ".config" / "outlook-cli"))
CONFIG_FILE = CONFIG_DIR / "config.yaml"
SIGNATURES_DIR = CONFIG_DIR / "signatures"

# Extended property for deferred/scheduled send
DEFERRED_SEND_PROPERTY_ID = "SystemTime 0x3FEF"

# Scheduled messages tracking
SCHEDULED_FILE = CACHE_DIR / "scheduled.json"



"""Manage email signatures for outlook-cli.

Signatures are stored as HTML files in ~/.config/outlook-cli/signatures/.
They can be pulled from sent emails or created manually.
"""

from __future__ import annotations

from pathlib import Path

import httpx

from .constants import BASE_URL, SIGNATURES_DIR
from .exceptions import ResourceNotFoundError


def list_signatures() -> list[str]:
    """Return names of saved signatures (without .html extension)."""
    if not SIGNATURES_DIR.exists():
        return []
    return sorted(p.stem for p in SIGNATURES_DIR.glob("*.html"))


def get_signature(name: str) -> str:
    """Load a signature's HTML content by name."""
    path = SIGNATURES_DIR / f"{name}.html"
    if not path.exists():
        raise ResourceNotFoundError(
            f"Signature '{name}' not found. Run 'outlook signature list' to see available signatures."
        )
    return path.read_text(encoding="utf-8")


def save_signature(name: str, html: str) -> Path:
    """Save signature HTML to disk."""
    SIGNATURES_DIR.mkdir(parents=True, exist_ok=True)
    path = SIGNATURES_DIR / f"{name}.html"
    path.write_text(html, encoding="utf-8")
    return path


def delete_signature(name: str) -> None:
    """Delete a saved signature."""
    path = SIGNATURES_DIR / f"{name}.html"
    if not path.exists():
        raise ResourceNotFoundError(f"Signature '{name}' not found.")
    path.unlink()


def pull_signature(token: str) -> tuple[str, str]:
    """Extract signature HTML from the most recent sent email that contains one.

    Looks for emails with structured signature blocks (containing phone/email/company info).

    Returns (signature_html, source_subject).
    """
    client = httpx.Client(
        base_url=BASE_URL,
        headers={"Authorization": f"Bearer {token}", "Content-Type": "application/json"},
        timeout=30,
    )
    try:
        for skip in range(0, 50, 5):
            resp = client.get(
                "/MailFolders/SentItems/messages",
                params={
                    "$top": 5,
                    "$skip": skip,
                    "$orderby": "SentDateTime desc",
                    "$select": "Subject,Body",
                },
            )
            if resp.status_code != 200:
                break
            for m in resp.json().get("value", []):
                body = m.get("Body", {}).get("Content", "")
                subject = m.get("Subject", "")
                sig = _extract_signature(body)
                if sig:
                    return sig, subject
    finally:
        client.close()
    raise ResourceNotFoundError(
        "Could not find a signature in your sent emails. "
        "Send an email with your signature from Outlook first."
    )


def _extract_signature(html: str) -> str | None:
    """Extract the signature table block from an email HTML body.

    Strategy: find all top-level <table> blocks before the reply chain,
    then pick the one containing mailto: links (signature indicator).
    The outermost table containing that block is the full signature.
    """
    # Cut content before reply chain
    reply_markers = ["divRplyFwdMsg", "x_divRplyFwdMsg", "gmail_quote", "originalMessage"]
    main_body = html
    for marker in reply_markers:
        idx = html.find(marker)
        if idx > 0:
            # Go back to find the enclosing div/hr
            cut = html.rfind("<hr", 0, idx)
            if cut < 0:
                cut = html.rfind("<div", 0, idx)
            main_body = html[:cut] if cut > 0 else html[:idx]
            break

    if "mailto:" not in main_body:
        return None

    # Find all <table starts in the main body
    table_starts = []
    pos = 0
    while True:
        idx = main_body.find("<table", pos)
        if idx < 0:
            break
        table_starts.append(idx)
        pos = idx + 1

    if not table_starts:
        return None

    # For each table start, extract the full table and check if it contains mailto:
    # Go from earliest to find the outermost signature container
    best = None
    for start in table_starts:
        # Extract balanced table
        table_html = _extract_balanced_table(main_body, start)
        if table_html and "mailto:" in table_html:
            # Prefer the outermost (earliest) table that contains mailto
            if best is None or start < best[0]:
                best = (start, table_html)

    if best is None:
        return None

    sig_html = best[1]

    # Validate: must be substantial
    if len(sig_html) < 100:
        return None

    return sig_html


def _extract_balanced_table(html: str, start: int) -> str | None:
    """Extract a complete <table>...</table> block with balanced nesting."""
    depth = 0
    i = start
    while i < len(html):
        if html[i:].startswith("<table"):
            depth += 1
            i += 6
        elif html[i:].startswith("</table>"):
            depth -= 1
            i += 8
            if depth == 0:
                return html[start:i]
        else:
            i += 1
    return None


def append_signature(body: str, signature_html: str, is_html: bool) -> tuple[str, bool]:
    """Append signature to email body.

    If body is plain text and signature is HTML, wraps body in HTML.
    Returns (new_body, is_html).
    """
    sig_block = f'<br><div id="Signature">{signature_html}<br></div>'

    if is_html:
        # Insert signature before closing </body> or </html> or at the end
        for tag in ["</body>", "</html>"]:
            idx = body.lower().rfind(tag)
            if idx > 0:
                return body[:idx] + sig_block + body[idx:], True
        return body + sig_block, True
    else:
        # Plain text body + HTML signature → convert to HTML
        escaped = body.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
        html_body = escaped.replace("\n", "<br>")
        full = (
            '<html><body style="font-family:Calibri,Arial,sans-serif;font-size:11pt;color:#000000;">'
            f"<div>{html_body}</div>"
            f"{sig_block}"
            "</body></html>"
        )
        return full, True



from __future__ import annotations

import json
from dataclasses import asdict
from datetime import datetime

from .models import Attachment, Contact, Email, Event, Folder

SCHEMA_VERSION = "1"


class _Encoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return super().default(o)


def _normalize(items):
    """Convert dataclasses / mixed lists to plain dicts."""
    if isinstance(items, list):
        return [asdict(i) if hasattr(i, "__dataclass_fields__") else i for i in items]
    if hasattr(items, "__dataclass_fields__"):
        return asdict(items)
    return items


def to_json(items: list | dict, pretty: bool = True) -> str:
    """Raw JSON — used by save_json for file export."""
    return json.dumps(_normalize(items), cls=_Encoder, indent=2 if pretty else None, ensure_ascii=False)


def to_json_envelope(items: list | dict, pretty: bool = True) -> str:
    """Wrap data in {ok, schema_version, data} envelope for stdout."""
    envelope = {
        "ok": True,
        "schema_version": SCHEMA_VERSION,
        "data": _normalize(items),
    }
    return json.dumps(envelope, cls=_Encoder, indent=2 if pretty else None, ensure_ascii=False)


def error_json(code: str, message: str) -> str:
    """Structured error envelope for --json mode."""
    envelope = {
        "ok": False,
        "schema_version": SCHEMA_VERSION,
        "error": {"code": code, "message": message},
    }
    return json.dumps(envelope, indent=2, ensure_ascii=False)


def save_json(items: list | dict, path: str) -> None:
    """Save raw JSON to file (no envelope — file export is raw data)."""
    with open(path, "w") as f:
        f.write(to_json(items))



"""Signature commands: signature-pull, signature-list, signature-show, signature-delete."""

from __future__ import annotations

import click

from ._common import (
    _handle_api_error,
    cfg,
    console,
    get_token,
    print_success,
)


@click.command("signature-pull")
@click.option("--name", "-n", default=None, help="Name for the signature (default: auto-detect)")
@_handle_api_error
def signature_pull(name: str | None):
    """Extract your signature from a recent sent email and save it."""
    from ..signature_manager import pull_signature, save_signature

    token = get_token()
    sig_html, source_subject = pull_signature(token)

    if not name:
        name = click.prompt("Signature name", default="default")

    path = save_signature(name, sig_html)
    print_success(f"Signature '{name}' saved from: {source_subject}")
    console.print(f"  [dim]{path}[/dim]")


@click.command("signature-list")
def signature_list():
    """List saved signatures."""
    from ..signature_manager import list_signatures

    sigs = list_signatures()
    if not sigs:
        print_success("No signatures saved. Run 'outlook signature-pull' to extract one.")
    else:
        for s in sigs:
            default = " [bold cyan](default)[/bold cyan]" if s == cfg.get("default_signature") else ""
            console.print(f"  {s}{default}")


@click.command("signature-show")
@click.argument("name")
@_handle_api_error
def signature_show(name: str):
    """Preview a saved signature."""
    from ..signature_manager import get_signature

    from bs4 import BeautifulSoup

    sig_html = get_signature(name)
    text = BeautifulSoup(sig_html, "html.parser").get_text("\n", strip=True)
    console.print(text)


@click.command("signature-delete")
@click.argument("name")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def signature_delete(name: str, yes: bool):
    """Delete a saved signature."""
    from ..signature_manager import delete_signature

    if not yes:
        click.confirm(f"Delete signature '{name}'?", abort=True)
    delete_signature(name)
    print_success(f"Deleted signature '{name}'")



"""Management commands: mark-read, move, delete."""

from __future__ import annotations

import click

from ._common import _get_client, _handle_api_error, print_success


@click.command("mark-read")
@click.argument("message_ids", nargs=-1, required=True)
@click.option("--unread", is_flag=True, help="Mark as unread instead")
@_handle_api_error
def mark_read(message_ids: tuple, unread: bool):
    """Mark messages as read (or unread with --unread). Accepts multiple IDs."""
    client = _get_client()
    status = "unread" if unread else "read"
    for mid in message_ids:
        client.mark_read(mid, is_read=not unread)
        print_success(f"Message #{mid} marked as {status}")


@click.command()
@click.argument("message_ids", nargs=-1, required=True)
@click.argument("destination")
@_handle_api_error
def move(message_ids: tuple, destination: str):
    """Move messages to another folder. Accepts multiple IDs."""
    client = _get_client()
    for mid in message_ids:
        client.move_message(mid, destination)
        print_success(f"Message #{mid} moved to {destination}")


@click.command()
@click.argument("message_ids", nargs=-1, required=True)
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def delete(message_ids: tuple, yes: bool):
    """Delete messages. Accepts multiple IDs."""
    if not yes:
        ids_str = ", ".join(f"#{m}" for m in message_ids)
        click.confirm(f"Delete {ids_str}?", abort=True)
    client = _get_client()
    for mid in message_ids:
        client.delete_message(mid)
        print_success(f"Message #{mid} deleted")



"""Manage Outlook master categories via OWA service.svc API.

Uses the UpdateMasterCategoryList action with a request-wrapped payload.
Key discovery: the payload must be wrapped in {"request": {...}} and sent
via x-owa-urlpostdata header (not in the body).

Operations:
  - Create: AddCategoryList
  - Delete: RemoveCategoryList
  - Rename: RemoveCategoryList + AddCategoryList (with same Id, new Name)
  - Recolor: ChangeCategoryColorList
"""

from __future__ import annotations

import json
import time
import uuid
from datetime import datetime, timezone
from typing import Callable
from urllib.parse import quote

import httpx

from .constants import BASE_URL, USER_AGENT
from .exceptions import ResourceNotFoundError, TokenExpiredError

OWA_SERVICE_BASE = "https://outlook.cloud.microsoft/owa/service.svc"


def _owa_request(token: str, action: str, payload: dict) -> dict:
    """Send OWA service.svc request with x-owa-urlpostdata pattern."""
    resp = httpx.post(
        f"{OWA_SERVICE_BASE}?action={action}",
        headers={
            "Authorization": f"Bearer {token}",
            "User-Agent": USER_AGENT,
            "Content-Type": "application/json; charset=utf-8",
            "Action": action,
            "x-req-source": "Mail",
            "x-owa-urlpostdata": quote(json.dumps(payload), safe=""),
        },
        content=b"",
        timeout=15,
    )
    if resp.status_code == 401:
        raise TokenExpiredError("Token expired. Run: outlook login")
    resp.raise_for_status()
    return resp.json()


def _update_master_categories(
    token: str,
    add: list[dict] | None = None,
    remove: list[str] | None = None,
    change_color: list[dict] | None = None,
) -> dict:
    """Call UpdateMasterCategoryList with the request-wrapped payload."""
    payload = {
        "request": {
            "__type": "UpdateMasterCategoryListRequest:#Exchange",
            "AddCategoryList": add or [],
            "RemoveCategoryList": remove or [],
            "ChangeCategoryColorList": change_color or [],
            "UpdateCategoryLastTimeUsedList": [],
            "ChangeCategoryKeyboardShortcutList": [],
        }
    }
    return _owa_request(token, "UpdateMasterCategoryList", payload)


def get_master_categories(token: str) -> list[dict]:
    """Fetch master category list via GetOwaUserConfiguration."""
    payload = {
        "__type": "GetOwaUserConfigurationJsonRequest:#Exchange",
        "Header": {
            "__type": "JsonRequestHeaders:#Exchange",
            "RequestServerVersion": "V2018_01_08",
        },
        "Body": {
            "__type": "GetOwaUserConfigurationRequest:#Exchange",
            "Owaconfigs": ["MasterCategoryList"],
        },
    }
    resp = _owa_request(token, "GetOwaUserConfiguration", payload)
    return resp.get("MasterCategoryList", {}).get("MasterList", [])


def create_category(token: str, name: str, color: int = 15) -> dict:
    """Create a new master category."""
    cat = {
        "Name": name,
        "Color": color,
        "Id": str(uuid.uuid4()),
        "LastTimeUsed": datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
        "KeyboardShortcut": 0,
    }
    return _update_master_categories(token, add=[cat])


def delete_category(token: str, name: str) -> dict:
    """Delete a master category by name."""
    return _update_master_categories(token, remove=[name])


def rename_category(
    token: str,
    old_name: str,
    new_name: str,
    propagate: bool = True,
    on_progress: Callable[[int, int], None] | None = None,
) -> int:
    """Rename a master category and optionally propagate to all messages.

    Returns the number of messages updated.
    """
    master = get_master_categories(token)
    existing = next((c for c in master if c["Name"] == old_name), None)
    if not existing:
        raise ResourceNotFoundError(f"Category '{old_name}' not found.")

    new_cat = {
        **existing,
        "Name": new_name,
        "LastTimeUsed": datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
    }
    _update_master_categories(token, add=[new_cat], remove=[old_name])

    if not propagate:
        return 0

    return _bulk_rename_on_messages(token, old_name, new_name, on_progress)


def _bulk_rename_on_messages(
    token: str,
    old_name: str,
    new_name: str,
    on_progress: Callable[[int, int], None] | None = None,
) -> int:
    """Rename a category label on all messages that have it."""
    client = httpx.Client(
        headers={"Authorization": f"Bearer {token}", "Content-Type": "application/json"},
        timeout=30,
    )
    total = 0
    try:
        while True:
            r = client.get(
                f"{BASE_URL}/messages",
                params={
                    "$top": 50,
                    "$filter": f"Categories/any(c:c eq '{old_name}')",
                    "$select": "Id,Categories",
                },
            )
            if r.status_code == 429:
                time.sleep(int(r.headers.get("Retry-After", 5)))
                continue
            if r.status_code != 200:
                break
            msgs = r.json().get("value", [])
            if not msgs:
                break
            for m in msgs:
                new_cats = [new_name if c == old_name else c for c in m["Categories"]]
                for attempt in range(3):
                    try:
                        r2 = client.patch(
                            f"{BASE_URL}/messages/{m['Id']}",
                            json={"Categories": new_cats},
                        )
                        if r2.status_code == 200:
                            total += 1
                            break
                        elif r2.status_code == 429:
                            time.sleep(int(r2.headers.get("Retry-After", 5)))
                        else:
                            time.sleep(2)
                    except httpx.ReadTimeout:
                        time.sleep(3)
            if on_progress:
                on_progress(total, -1)
    finally:
        client.close()
    return total


def clear_category(
    token: str,
    name: str,
    folder: str | None = None,
    max_messages: int | None = None,
    on_progress: Callable[[int, int], None] | None = None,
) -> int:
    """Remove a category label from messages. Does not touch master category list.

    Args:
        folder: Limit to a specific folder (e.g. "Inbox"). None = all folders.
        max_messages: Stop after clearing this many messages. None = all.

    Returns the number of messages updated.
    """
    client = httpx.Client(
        headers={"Authorization": f"Bearer {token}", "Content-Type": "application/json"},
        timeout=30,
    )
    total = 0
    try:
        while True:
            if folder:
                url = f"{BASE_URL}/MailFolders/{folder}/messages"
            else:
                url = f"{BASE_URL}/messages"
            r = client.get(
                url,
                params={
                    "$top": 50,
                    "$filter": f"Categories/any(c:c eq '{name}')",
                    "$select": "Id,Categories",
                },
            )
            if r.status_code == 429:
                time.sleep(int(r.headers.get("Retry-After", 5)))
                continue
            if r.status_code != 200:
                break
            msgs = r.json().get("value", [])
            if not msgs:
                break
            for m in msgs:
                new_cats = [c for c in m["Categories"] if c != name]
                for attempt in range(3):
                    try:
                        r2 = client.patch(
                            f"{BASE_URL}/messages/{m['Id']}",
                            json={"Categories": new_cats},
                        )
                        if r2.status_code == 200:
                            total += 1
                            break
                        elif r2.status_code == 429:
                            time.sleep(int(r2.headers.get("Retry-After", 5)))
                        else:
                            time.sleep(2)
                    except httpx.ReadTimeout:
                        time.sleep(3)
                if max_messages and total >= max_messages:
                    break
            if on_progress:
                on_progress(total, -1)
            if max_messages and total >= max_messages:
                break
    finally:
        client.close()
    return total


def recolor_category(token: str, name: str, color: int) -> dict:
    """Change a master category's color."""
    return _update_master_categories(
        token,
        change_color=[{"Name": name, "Color": color}],
    )



from __future__ import annotations

import json
import os
import stat
import time
from base64 import urlsafe_b64decode
from pathlib import Path

import httpx

from .constants import BASE_URL, BROWSER_STATE_FILE, CACHE_DIR, OWA_URL, TOKEN_FILE, USER_AGENT
from .exceptions import AuthRequiredError


def get_token() -> str:
    """Return a valid bearer token, from env, cache, or interactive login."""
    # 1. Environment variable
    env_token = os.environ.get("OUTLOOK_TOKEN")
    if env_token:
        return env_token

    # 2. Cached token
    cached = _load_cached_token()
    if cached:
        return cached

    # 3. Interactive login
    return login()


def login(force: bool = False, debug: bool = False) -> str:
    """Launch Playwright browser to capture bearer token from OWA."""
    from playwright.sync_api import sync_playwright

    CACHE_DIR.mkdir(parents=True, exist_ok=True)

    captured_token: list[str] = []
    seen_urls: list[str] = []

    def _intercept_request(request):
        auth = request.headers.get("authorization", "")
        if auth.lower().startswith("bearer "):
            token = auth.split(" ", 1)[1]
            if debug:
                seen_urls.append(request.url[:120])
                print(f"  [debug] Bearer token in: {request.url[:120]}")
            if len(token) > 100:
                captured_token.append(token)
                if debug:
                    print(f"  [debug] Captured token ({len(token)} chars)")

    with sync_playwright() as p:
        launch_args: dict = {}
        if BROWSER_STATE_FILE.exists() and not force:
            launch_args["storage_state"] = str(BROWSER_STATE_FILE)

        browser = p.chromium.launch(headless=False)
        context = browser.new_context(
            user_agent=USER_AGENT,
            **launch_args,
        )

        # Listen on CONTEXT level to catch all frames/workers
        context.on("request", _intercept_request)

        page = context.new_page()

        print("Opening Outlook... Log in and wait for your inbox to load.")
        print("The browser will close automatically once the token is captured.")
        page.goto(OWA_URL, wait_until="domcontentloaded")

        # Poll: wait up to 120s for a token to appear
        deadline = time.time() + 120
        while not captured_token and time.time() < deadline:
            try:
                page.wait_for_timeout(2000)
            except Exception:
                # Browser was closed by user
                break

            # After 20s without token, try triggering an API call
            if not captured_token and time.time() > deadline - 95:
                try:
                    page.evaluate("""
                        fetch('/api/v2.0/me', {credentials: 'include'})
                            .catch(() => {});
                    """)
                except Exception:
                    pass

        # Save browser state for future SSO
        try:
            context.storage_state(path=str(BROWSER_STATE_FILE))
            _chmod_600(BROWSER_STATE_FILE)
        except Exception:
            pass

        try:
            browser.close()
        except Exception:
            pass

    if debug and seen_urls:
        print(f"\n  [debug] Total requests with Bearer: {len(seen_urls)}")

    if not captured_token:
        raise AuthRequiredError(
            "Could not capture bearer token.\n"
            "Make sure you logged in and your inbox fully loaded.\n"
            "Tip: Try 'outlook login --debug' to see request details."
        )

    # Deduplicate and pick the best token
    unique_tokens = list(dict.fromkeys(captured_token))  # preserve order, remove dupes
    token = _pick_best_token(unique_tokens, debug=debug)
    _save_token(token)
    return token


def _pick_best_token(tokens: list[str], debug: bool = False) -> str:
    """Try each token against known endpoints. Prefer one that can read mail."""
    import httpx

    # Group by decoded audience
    candidates: list[tuple[str, str]] = []  # (token, audience)
    for t in tokens:
        aud = _decode_audience(t)
        candidates.append((t, aud))

    if debug:
        for t, aud in candidates:
            print(f"  [debug] Token ({len(t)} chars) audience={aud}")

    # Try each token: first against REST v2, then Graph
    endpoints = [
        ("https://outlook.office.com/api/v2.0/me/messages?$top=1", "REST v2"),
        ("https://outlook.office365.com/api/v2.0/me/messages?$top=1", "REST v2 (365)"),
        ("https://graph.microsoft.com/v1.0/me/messages?$top=1", "Graph"),
    ]

    for token, aud in candidates:
        for url, label in endpoints:
            try:
                resp = httpx.get(
                    url,
                    headers={"Authorization": f"Bearer {token}", "User-Agent": USER_AGENT},
                    timeout=10,
                )
                if resp.status_code == 200:
                    if debug:
                        print(f"  [debug] Token works with {label}!")
                    return token
            except httpx.HTTPError:
                continue

    # If no token works for mail, try /me as fallback
    for token, aud in candidates:
        for base in ("https://outlook.office.com/api/v2.0", "https://graph.microsoft.com/v1.0"):
            try:
                resp = httpx.get(
                    f"{base}/me",
                    headers={"Authorization": f"Bearer {token}", "User-Agent": USER_AGENT},
                    timeout=10,
                )
                if resp.status_code == 200:
                    if debug:
                        print(f"  [debug] Token works for /me at {base} (no mail access though)")
                    return token
            except httpx.HTTPError:
                continue

    # Last resort: return the longest token (likely the OWA one)
    return max(tokens, key=len)


def _decode_audience(token: str) -> str:
    try:
        parts = token.split(".")
        if len(parts) < 2:
            return "unknown"
        payload = parts[1]
        payload += "=" * (4 - len(payload) % 4)
        decoded = json.loads(urlsafe_b64decode(payload))
        return decoded.get("aud", "unknown")
    except Exception:
        return "unknown"


def verify_token(token: str) -> bool:
    """Check if token is valid by calling /me endpoint."""
    try:
        resp = httpx.get(
            f"{BASE_URL}",
            headers={"Authorization": f"Bearer {token}", "User-Agent": USER_AGENT},
            timeout=10,
        )
        return resp.status_code == 200
    except httpx.HTTPError:
        return False


def _load_cached_token() -> str | None:
    if not TOKEN_FILE.exists():
        return None
    try:
        data = json.loads(TOKEN_FILE.read_text())
        token = data["token"]
        expires_at = data.get("expires_at", 0)
        # Check expiry with 5-minute buffer
        if time.time() > expires_at - 300:
            return None
        return token
    except (json.JSONDecodeError, KeyError):
        return None


def _save_token(token: str) -> None:
    CACHE_DIR.mkdir(parents=True, exist_ok=True)
    expires_at = _decode_exp(token)
    data = {"token": token, "expires_at": expires_at}
    TOKEN_FILE.write_text(json.dumps(data))
    _chmod_600(TOKEN_FILE)


def _decode_exp(token: str) -> float:
    """Extract exp claim from JWT without full verification."""
    try:
        parts = token.split(".")
        if len(parts) < 2:
            return time.time() + 3600  # fallback 1h
        payload = parts[1]
        # Fix padding
        payload += "=" * (4 - len(payload) % 4)
        decoded = json.loads(urlsafe_b64decode(payload))
        return float(decoded.get("exp", time.time() + 3600))
    except Exception:
        return time.time() + 3600


def _chmod_600(path: Path) -> None:
    try:
        path.chmod(stat.S_IRUSR | stat.S_IWUSR)
    except OSError:
        pass



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime


@dataclass
class EmailAddress:
    name: str
    address: str

    @classmethod
    def from_api(cls, data: dict) -> EmailAddress:
        ea = data.get("EmailAddress", data)
        return cls(name=ea.get("Name", ""), address=ea.get("Address", ""))

    def __str__(self) -> str:
        if self.name:
            return f"{self.name} <{self.address}>"
        return self.address


@dataclass
class Email:
    id: str
    subject: str
    sender: EmailAddress
    to: list[EmailAddress]
    cc: list[EmailAddress]
    received: datetime
    preview: str
    body: str
    body_type: str  # "Text" or "HTML"
    is_read: bool
    has_attachments: bool
    importance: str
    conversation_id: str
    categories: list[str] = field(default_factory=list)
    scheduled_send: datetime | None = None
    # display number assigned by CLI
    display_num: int = 0

    @classmethod
    def from_api(cls, data: dict) -> Email:
        scheduled_send = None
        for prop in data.get("SingleValueExtendedProperties", []):
            if "0x3FEF" in prop.get("PropertyId", ""):
                scheduled_send = _parse_dt(prop.get("Value", ""))
                break

        return cls(
            id=data["Id"],
            subject=data.get("Subject", "(No Subject)"),
            sender=EmailAddress.from_api(data.get("From", {}).get("EmailAddress", {"Name": "", "Address": ""})),
            to=[EmailAddress.from_api(r) for r in data.get("ToRecipients", [])],
            cc=[EmailAddress.from_api(r) for r in data.get("CcRecipients", [])],
            received=_parse_dt(data.get("ReceivedDateTime", "")),
            preview=data.get("BodyPreview", ""),
            body=data.get("Body", {}).get("Content", ""),
            body_type=data.get("Body", {}).get("ContentType", "Text"),
            is_read=data.get("IsRead", False),
            has_attachments=data.get("HasAttachments", False),
            importance=data.get("Importance", "Normal"),
            conversation_id=data.get("ConversationId", ""),
            categories=data.get("Categories", []),
            scheduled_send=scheduled_send,
        )


@dataclass
class Folder:
    id: str
    name: str
    unread_count: int
    total_count: int
    parent_folder_id: str

    @classmethod
    def from_api(cls, data: dict) -> Folder:
        return cls(
            id=data["Id"],
            name=data.get("DisplayName", ""),
            unread_count=data.get("UnreadItemCount", 0),
            total_count=data.get("TotalItemCount", 0),
            parent_folder_id=data.get("ParentFolderId", ""),
        )


@dataclass
class Attachment:
    id: str
    name: str
    content_type: str
    size: int
    is_inline: bool
    content_bytes: str | None = None  # base64

    @classmethod
    def from_api(cls, data: dict) -> Attachment:
        return cls(
            id=data["Id"],
            name=data.get("Name", ""),
            content_type=data.get("ContentType", ""),
            size=data.get("Size", 0),
            is_inline=data.get("IsInline", False),
            content_bytes=data.get("ContentBytes"),
        )


@dataclass
class Attendee:
    email: EmailAddress
    type: str  # "Required", "Optional", "Resource"
    response: str  # "None", "Accepted", "Declined", "TentativelyAccepted", "NotResponded"

    @classmethod
    def from_api(cls, data: dict) -> Attendee:
        return cls(
            email=EmailAddress.from_api(data.get("EmailAddress", {})),
            type=data.get("Type", "Required"),
            response=data.get("Status", {}).get("Response", "None"),
        )


@dataclass
class Event:
    id: str
    subject: str
    start: datetime
    end: datetime
    location: str
    organizer: EmailAddress
    is_all_day: bool
    body_preview: str
    body: str
    body_type: str
    attendees: list[Attendee] = field(default_factory=list)
    categories: list[str] = field(default_factory=list)
    show_as: str = "Busy"
    sensitivity: str = "Normal"
    is_cancelled: bool = False
    response_status: str = ""
    web_link: str = ""
    is_online_meeting: bool = False
    online_meeting_url: str = ""
    recurrence: dict | None = None
    event_type: str = ""  # "SingleInstance", "Occurrence", "Exception", "SeriesMaster"
    series_master_id: str = ""
    display_num: int = 0

    @classmethod
    def from_api(cls, data: dict) -> Event:
        org_data = data.get("Organizer", {}).get("EmailAddress", {"Name": "", "Address": ""})
        attendees = [Attendee.from_api(a) for a in data.get("Attendees", [])]
        online_url = ""
        if data.get("OnlineMeeting"):
            online_url = data["OnlineMeeting"].get("JoinUrl", "")
        return cls(
            id=data["Id"],
            subject=data.get("Subject", "(No Subject)"),
            start=_parse_dt(data.get("Start", {}).get("DateTime", "")),
            end=_parse_dt(data.get("End", {}).get("DateTime", "")),
            location=data.get("Location", {}).get("DisplayName", ""),
            organizer=EmailAddress.from_api(org_data),
            is_all_day=data.get("IsAllDay", False),
            body_preview=data.get("BodyPreview", ""),
            body=data.get("Body", {}).get("Content", ""),
            body_type=data.get("Body", {}).get("ContentType", "Text"),
            attendees=attendees,
            categories=data.get("Categories", []),
            show_as=data.get("ShowAs", "Busy"),
            sensitivity=data.get("Sensitivity", "Normal"),
            is_cancelled=data.get("IsCancelled", False),
            response_status=data.get("ResponseStatus", {}).get("Response", ""),
            web_link=data.get("WebLink", ""),
            is_online_meeting=data.get("IsOnlineMeeting", False),
            online_meeting_url=online_url,
            recurrence=data.get("Recurrence"),
            event_type=data.get("Type", ""),
            series_master_id=data.get("SeriesMasterId", ""),
        )


@dataclass
class Contact:
    id: str
    display_name: str
    given_name: str
    surname: str
    email_addresses: list[EmailAddress]
    company: str
    job_title: str

    @classmethod
    def from_api(cls, data: dict) -> Contact:
        emails = [
            EmailAddress(name=e.get("Name", ""), address=e.get("Address", ""))
            for e in data.get("EmailAddresses", [])
        ]
        return cls(
            id=data["Id"],
            display_name=data.get("DisplayName", ""),
            given_name=data.get("GivenName", ""),
            surname=data.get("Surname", ""),
            email_addresses=emails,
            company=data.get("CompanyName", ""),
            job_title=data.get("JobTitle", ""),
        )


def _parse_dt(s: str) -> datetime:
    if not s:
        return datetime.min
    # Outlook returns ISO 8601 with or without timezone
    s = s.replace("Z", "+00:00")
    try:
        return datetime.fromisoformat(s)
    except ValueError:
        return datetime.min



"""Tests for serialization.py — envelope format, error_json, encoding."""

from __future__ import annotations

import json
import tempfile
from datetime import datetime, timezone

from outlook_cli.models import Email, EmailAddress, Folder
from outlook_cli.serialization import error_json, save_json, to_json, to_json_envelope


class TestToJsonEnvelope:
    def test_success_envelope_structure(self):
        result = json.loads(to_json_envelope({"key": "value"}))
        assert result["ok"] is True
        assert result["schema_version"] == "1"
        assert result["data"] == {"key": "value"}

    def test_envelope_with_list(self):
        result = json.loads(to_json_envelope([1, 2, 3]))
        assert result["ok"] is True
        assert result["data"] == [1, 2, 3]

    def test_envelope_with_dataclass(self):
        folder = Folder(id="f1", name="Inbox", unread_count=5, total_count=100, parent_folder_id="root")
        result = json.loads(to_json_envelope(folder))
        assert result["ok"] is True
        assert result["data"]["name"] == "Inbox"
        assert result["data"]["unread_count"] == 5

    def test_envelope_with_dataclass_list(self):
        folders = [
            Folder(id="f1", name="Inbox", unread_count=5, total_count=100, parent_folder_id="root"),
            Folder(id="f2", name="Sent", unread_count=0, total_count=50, parent_folder_id="root"),
        ]
        result = json.loads(to_json_envelope(folders))
        assert result["ok"] is True
        assert len(result["data"]) == 2
        assert result["data"][1]["name"] == "Sent"


class TestErrorJson:
    def test_error_envelope_structure(self):
        result = json.loads(error_json("not_found", "Message #999 not found"))
        assert result["ok"] is False
        assert result["schema_version"] == "1"
        assert result["error"]["code"] == "not_found"
        assert result["error"]["message"] == "Message #999 not found"

    def test_error_codes(self):
        for code in ["session_expired", "rate_limited", "not_found", "not_authenticated", "unknown_error"]:
            result = json.loads(error_json(code, "test"))
            assert result["error"]["code"] == code


class TestToJsonRaw:
    def test_no_envelope(self):
        """to_json() should NOT include envelope — it's for file export."""
        result = json.loads(to_json({"subject": "Hello"}))
        assert "ok" not in result
        assert result["subject"] == "Hello"

    def test_datetime_encoding(self):
        dt = datetime(2026, 3, 15, 10, 0, 0, tzinfo=timezone.utc)
        result = json.loads(to_json({"when": dt}))
        assert "2026-03-15" in result["when"]


class TestSaveJson:
    def test_writes_raw_to_file(self):
        with tempfile.NamedTemporaryFile(suffix=".json", delete=False, mode="r") as f:
            path = f.name
        save_json({"test": True}, path)
        data = json.loads(open(path).read())
        assert "ok" not in data  # no envelope
        assert data["test"] is True



"""Contact commands: contacts."""

from __future__ import annotations

import click

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    print_contacts,
    print_success,
    save_json,
    to_json_envelope,
)


@click.command()
@click.option("--max", "-n", "max_count", default=50, type=int, help="Max contacts")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--output", "-o", type=click.Path(), help="Save output to file")
@_handle_api_error
def contacts(max_count: int, as_json: bool, output: str | None):
    """List contacts."""
    client = _get_client()
    contact_list = client.get_contacts(top=max_count)

    if _wants_json(as_json):
        if output:
            save_json(contact_list, output)
            print_success(f"Saved to {output}")
        else:
            click.echo(to_json_envelope(contact_list))
    else:
        if not contact_list:
            print_success("No contacts found.")
        else:
            print_contacts(contact_list)



"""Outlook 365 CLI entry point — command registration only."""

from __future__ import annotations

import click

from .commands import (
    attachments as attachments_mod,
    auth as auth_mod,
    calendar as calendar_mod,
    categories as categories_mod,
    contacts as contacts_mod,
    folders as folders_mod,
    mail as mail_mod,
    manage as manage_mod,
    schedule as schedule_mod,
    search as search_mod,
    signatures as signatures_mod,
)


@click.group()
@click.version_option(package_name="outlook-cli")
def cli():
    """Outlook 365 CLI - read, send, and manage emails from the terminal."""
    pass


# Auth
cli.add_command(auth_mod.login)
cli.add_command(auth_mod.whoami)

# Mail - Read & Write
cli.add_command(mail_mod.inbox)
cli.add_command(mail_mod.read)
cli.add_command(mail_mod.thread)
cli.add_command(mail_mod.send)
cli.add_command(mail_mod.draft)
cli.add_command(mail_mod.draft_send)
cli.add_command(mail_mod.reply)
cli.add_command(mail_mod.reply_draft)
cli.add_command(mail_mod.forward)

# Scheduled send
cli.add_command(schedule_mod.schedule)
cli.add_command(schedule_mod.schedule_list)
cli.add_command(schedule_mod.schedule_cancel)
cli.add_command(schedule_mod.schedule_draft)

# Search
cli.add_command(search_mod.search)

# Folders
cli.add_command(folders_mod.folders)
cli.add_command(folders_mod.folder)

# Categories
cli.add_command(categories_mod.categories)
cli.add_command(categories_mod.categorize)
cli.add_command(categories_mod.uncategorize)
cli.add_command(categories_mod.category_rename)
cli.add_command(categories_mod.category_clear)
cli.add_command(categories_mod.category_delete)
cli.add_command(categories_mod.category_create)

# Signatures
cli.add_command(signatures_mod.signature_pull)
cli.add_command(signatures_mod.signature_list)
cli.add_command(signatures_mod.signature_show)
cli.add_command(signatures_mod.signature_delete)

# Management
cli.add_command(manage_mod.mark_read)
cli.add_command(manage_mod.move)
cli.add_command(manage_mod.delete)

# Attachments
cli.add_command(attachments_mod.attachments)

# Calendar
cli.add_command(calendar_mod.calendar)
cli.add_command(calendar_mod.event)
cli.add_command(calendar_mod.event_create)
cli.add_command(calendar_mod.event_update)
cli.add_command(calendar_mod.event_delete)
cli.add_command(calendar_mod.event_instances)
cli.add_command(calendar_mod.event_respond)
cli.add_command(calendar_mod.calendars_cmd)
cli.add_command(calendar_mod.free_busy)
cli.add_command(calendar_mod.people_search)

# Contacts
cli.add_command(contacts_mod.contacts)



"""Calendar commands: calendar, event, event-create/update/delete/instances/respond, calendars, free-busy, people-search."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone

import click

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    cfg,
    console,
    print_calendars,
    print_error,
    print_event_detail,
    print_events,
    print_meeting_suggestions,
    print_people,
    print_success,
    save_json,
    to_json_envelope,
)


def _parse_event_time(s: str) -> str:
    """Parse event time to ISO format for the API.

    Supports:
      +2h, +30m, +1h30m           relative offset
      tomorrow 09:00              relative day
      today 17:00                 relative day
      2026-03-15T10:00            ISO format
      2026-03-15 10:00            space-separated
    """
    import re
    now = datetime.now()

    # Relative offset: +30m, +1h, +2h30m
    offset_match = re.match(r'^\+(?:(\d+)h)?(?:(\d+)m)?$', s)
    if offset_match:
        hours = int(offset_match.group(1) or 0)
        minutes = int(offset_match.group(2) or 0)
        if hours == 0 and minutes == 0:
            raise click.BadParameter(f"Invalid offset: {s}")
        target = now + timedelta(hours=hours, minutes=minutes)
        return target.strftime("%Y-%m-%dT%H:%M:%S")

    # today/tomorrow HH:MM
    day_match = re.match(r'^(today|tomorrow)\s+(\d{1,2}:\d{2})$', s, re.IGNORECASE)
    if day_match:
        day_word, time_str = day_match.groups()
        h, m = map(int, time_str.split(":"))
        target = now.replace(hour=h, minute=m, second=0, microsecond=0)
        if day_word.lower() == "tomorrow":
            target += timedelta(days=1)
        return target.strftime("%Y-%m-%dT%H:%M:%S")

    # ISO-like: 2026-03-15T10:00 or 2026-03-15 10:00
    s_norm = s.replace(" ", "T", 1) if " " in s and "T" not in s else s
    try:
        dt = datetime.fromisoformat(s_norm)
        return dt.strftime("%Y-%m-%dT%H:%M:%S")
    except ValueError:
        pass

    raise click.BadParameter(
        f"Cannot parse '{s}'. Use: +1h, tomorrow 09:00, or 2026-03-15T10:00"
    )


def _build_recurrence(
    repeat: str,
    start_dt: str,
    interval: int = 1,
    count: int | None = None,
    until: str | None = None,
    days: str | None = None,
) -> dict:
    """Build Outlook API Recurrence payload."""
    start_date = start_dt[:10]  # YYYY-MM-DD
    day_of_week = datetime.fromisoformat(start_dt).strftime("%A")

    if repeat == "daily":
        pattern = {"Type": "Daily", "Interval": interval}
    elif repeat == "weekly":
        if days:
            day_list = [d.strip() for d in days.split(",")]
        else:
            day_list = [day_of_week]
        pattern = {"Type": "Weekly", "Interval": interval, "DaysOfWeek": day_list}
    elif repeat == "monthly":
        day_of_month = int(start_dt[8:10])
        pattern = {"Type": "AbsoluteMonthly", "Interval": interval, "DayOfMonth": day_of_month}
    else:
        raise click.BadParameter(f"Unknown repeat type: {repeat}")

    if count:
        rng = {"Type": "Numbered", "StartDate": start_date, "NumberOfOccurrences": count}
    elif until:
        rng = {"Type": "EndDate", "StartDate": start_date, "EndDate": until}
    else:
        rng = {"Type": "Numbered", "StartDate": start_date, "NumberOfOccurrences": 4}

    return {"Pattern": pattern, "Range": rng}


@click.command()
@click.option("--days", default=7, type=int, help="Number of days to show")
@click.option("--calendar", "cal_name", default=None, help="Calendar name (default: your primary calendar)")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--output", "-o", type=click.Path(), help="Save output to file")
@_handle_api_error
def calendar(days: int, cal_name: str | None, as_json: bool, output: str | None):
    """Show upcoming calendar events."""
    client = _get_client()
    now = datetime.now(timezone.utc)
    end = now + timedelta(days=days)
    events = client.get_calendar_view(
        start=now.isoformat(),
        end=end.isoformat(),
        calendar_name=cal_name,
    )

    if _wants_json(as_json):
        if output:
            save_json(events, output)
            print_success(f"Saved to {output}")
        else:
            click.echo(to_json_envelope(events))
    else:
        if not events:
            print_success(f"No events in the next {days} days.")
        else:
            console.print(f"[bold cyan]Calendar ({days} days)[/bold cyan]")
            print_events(events)


@click.command()
@click.argument("event_id")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def event(event_id: str, as_json: bool):
    """View event details by display number."""
    client = _get_client()
    ev = client.get_event(event_id)
    if _wants_json(as_json):
        click.echo(to_json_envelope(ev))
    else:
        print_event_detail(ev)


@click.command("event-create")
@click.argument("subject")
@click.argument("start")
@click.argument("end")
@click.option("--attendee", "-a", multiple=True, help="Attendee email (repeatable)")
@click.option("--location", "-l", default=None, help="Event location")
@click.option("--body", "-b", default=None, help="Event body/description")
@click.option("--html", "is_html", is_flag=True, help="Body is HTML")
@click.option("--all-day", is_flag=True, help="All-day event")
@click.option("--reminder", type=int, default=15, help="Reminder minutes before (default 15)")
@click.option("--teams", is_flag=True, help="Create as Teams online meeting")
@click.option("--repeat", type=click.Choice(["daily", "weekly", "monthly"]), default=None, help="Recurrence pattern")
@click.option("--repeat-interval", type=int, default=1, help="Repeat every N days/weeks/months (default 1)")
@click.option("--repeat-count", type=int, default=None, help="Number of occurrences")
@click.option("--repeat-until", default=None, help="End date for recurrence (YYYY-MM-DD)")
@click.option("--repeat-days", default=None, help="Days of week for weekly (comma-separated: Monday,Wednesday,Friday)")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def event_create(
    subject: str, start: str, end: str,
    attendee: tuple, location: str | None, body: str | None,
    is_html: bool, all_day: bool, reminder: int, teams: bool,
    repeat: str | None, repeat_interval: int, repeat_count: int | None,
    repeat_until: str | None, repeat_days: str | None,
    as_json: bool, yes: bool,
):
    """Create a calendar event.

    START and END accept: +1h, tomorrow 09:00, 2026-03-15T10:00

    Recurrence examples:
      --repeat weekly --repeat-count 4
      --repeat daily --repeat-until 2026-04-01
      --repeat weekly --repeat-days Monday,Wednesday --repeat-count 8
      --repeat monthly --repeat-count 3
    """
    start_dt = _parse_event_time(start)
    end_dt = _parse_event_time(end)
    attendees = list(attendee) if attendee else None

    recurrence = None
    if repeat:
        recurrence = _build_recurrence(
            repeat, start_dt, interval=repeat_interval,
            count=repeat_count, until=repeat_until, days=repeat_days,
        )

    if not yes:
        console.print(f"  [bold]Subject:[/bold] {subject}")
        console.print(f"  [bold]Start:[/bold] {start_dt}")
        console.print(f"  [bold]End:[/bold] {end_dt}")
        if attendees:
            console.print(f"  [bold]Attendees:[/bold] {', '.join(attendees)}")
        if location:
            console.print(f"  [bold]Location:[/bold] {location}")
        if recurrence:
            pat = recurrence["Pattern"]
            rng = recurrence["Range"]
            console.print(f"  [bold]Repeat:[/bold] {pat['Type']} every {pat['Interval']}")
            if rng["Type"] == "Numbered":
                console.print(f"  [bold]Occurrences:[/bold] {rng['NumberOfOccurrences']}")
            elif rng["Type"] == "EndDate":
                console.print(f"  [bold]Until:[/bold] {rng['EndDate']}")
        click.confirm("Create this event?", abort=True)

    client = _get_client()
    ev = client.create_event(
        subject=subject, start=start_dt, end=end_dt,
        timezone=cfg.get("timezone", "UTC"),
        attendees=attendees, location=location,
        body=body, html=is_html, is_all_day=all_day,
        reminder_minutes=reminder, is_online_meeting=teams,
        recurrence=recurrence,
    )

    if _wants_json(as_json):
        click.echo(to_json_envelope(ev))
    else:
        print_success(f"Event created: {ev.subject}")
        console.print(f"  [dim]{ev.start.strftime('%Y-%m-%d %H:%M')} - {ev.end.strftime('%H:%M')}[/dim]")
        if ev.attendees:
            console.print(f"  [dim]Attendees: {len(ev.attendees)}[/dim]")
        if ev.recurrence:
            from ..formatter import _format_recurrence
            console.print(f"  [dim]Recurrence: {_format_recurrence(ev.recurrence)}[/dim]")


@click.command("event-update")
@click.argument("event_id")
@click.option("--subject", "-s", default=None, help="New subject")
@click.option("--start", default=None, help="New start time")
@click.option("--end", default=None, help="New end time")
@click.option("--location", "-l", default=None, help="New location")
@click.option("--body", "-b", default=None, help="New body/description")
@click.option("--add-attendee", multiple=True, help="Add attendee email (repeatable)")
@click.option("--remove-attendee", multiple=True, help="Remove attendee email (repeatable)")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def event_update(
    event_id: str, subject: str | None, start: str | None, end: str | None,
    location: str | None, body: str | None,
    add_attendee: tuple, remove_attendee: tuple, as_json: bool,
):
    """Update a calendar event."""
    client = _get_client()

    if add_attendee:
        client.add_event_attendees(event_id, list(add_attendee))
        print_success(f"Added {len(add_attendee)} attendee(s) to event #{event_id}")
    if remove_attendee:
        client.remove_event_attendees(event_id, list(remove_attendee))
        print_success(f"Removed {len(remove_attendee)} attendee(s) from event #{event_id}")

    kwargs: dict = {}
    if subject:
        kwargs["subject"] = subject
    if start:
        kwargs["start"] = _parse_event_time(start)
    if end:
        kwargs["end"] = _parse_event_time(end)
    if location:
        kwargs["location"] = location
    if body:
        kwargs["body"] = body

    if kwargs:
        kwargs["timezone"] = cfg.get("timezone", "UTC")
        ev = client.update_event(event_id, **kwargs)
        if _wants_json(as_json):
            click.echo(to_json_envelope(ev))
        else:
            print_success(f"Event #{event_id} updated: {ev.subject}")
    elif not add_attendee and not remove_attendee:
        print_error("No changes specified. Use --subject, --start, --end, --location, --body, --add-attendee, --remove-attendee.")


@click.command("event-delete")
@click.argument("event_ids", nargs=-1, required=True)
@click.option("--series", is_flag=True, help="Delete entire recurring series (uses SeriesMasterId)")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def event_delete(event_ids: tuple, series: bool, yes: bool):
    """Delete calendar events. Accepts multiple IDs.

    For recurring events: deletes single occurrence by default.
    Use --series to delete the entire series.
    """
    client = _get_client()
    for eid in event_ids:
        if series:
            ev = client.get_event(eid)
            if ev.series_master_id:
                target_id = ev.series_master_id
                label = f"entire series of #{eid}"
            elif ev.event_type == "SeriesMaster":
                target_id = ev.id
                label = f"series #{eid}"
            else:
                target_id = ev.id
                label = f"event #{eid} (not a recurring event)"
            if not yes:
                click.confirm(f"Delete {label}?", abort=True)
            client._delete(f"/events/{target_id}")
            print_success(f"Deleted {label}")
        else:
            if not yes:
                click.confirm(f"Delete event #{eid}?", abort=True)
            client.delete_event(eid)
            print_success(f"Event #{eid} deleted")


@click.command("event-instances")
@click.argument("event_id")
@click.option("--days", default=90, type=int, help="Look-ahead days (default 90)")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def event_instances(event_id: str, days: int, as_json: bool):
    """List occurrences of a recurring event."""
    client = _get_client()
    now = datetime.now(timezone.utc)
    end = now + timedelta(days=days)
    events = client.get_event_instances(
        event_id,
        start=now.isoformat(),
        end=end.isoformat(),
    )
    if _wants_json(as_json):
        click.echo(to_json_envelope(events))
    else:
        if not events:
            print_success("No occurrences found.")
        else:
            console.print(f"[bold cyan]Occurrences ({len(events)})[/bold cyan]")
            print_events(events)


@click.command("event-respond")
@click.argument("event_id")
@click.argument("response", type=click.Choice(["accept", "decline", "tentative"]))
@click.option("--comment", "-c", default="", help="Response comment")
@click.option("--silent", is_flag=True, help="Don't send response to organizer")
@_handle_api_error
def event_respond(event_id: str, response: str, comment: str, silent: bool):
    """Respond to a meeting invitation (accept/decline/tentative)."""
    response_map = {
        "accept": "accept",
        "decline": "decline",
        "tentative": "tentativelyaccept",
    }
    client = _get_client()
    client.respond_to_event(
        event_id, response_map[response],
        comment=comment, send_response=not silent,
    )
    print_success(f"Event #{event_id}: {response}")


@click.command(name="calendars")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def calendars_cmd(as_json: bool):
    """List available calendars."""
    client = _get_client()
    cals = client.get_calendars()
    if _wants_json(as_json):
        click.echo(to_json_envelope(cals))
    else:
        if not cals:
            print_success("No calendars found.")
        else:
            print_calendars(cals)


@click.command("free-busy")
@click.argument("attendees")
@click.argument("date")
@click.option("--start-hour", default=9, type=int, help="Start hour (default 9)")
@click.option("--end-hour", default=18, type=int, help="End hour (default 18)")
@click.option("--duration", "-d", default=60, type=int, help="Meeting duration in minutes (default 60)")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def free_busy(attendees: str, date: str, start_hour: int, end_hour: int, duration: int, as_json: bool):
    """Find available meeting times.

    ATTENDEES: comma-separated emails. DATE: YYYY-MM-DD, today, or tomorrow.
    """
    addr_list = [a.strip() for a in attendees.split(",")]

    if date.lower() == "today":
        d = datetime.now()
    elif date.lower() == "tomorrow":
        d = datetime.now() + timedelta(days=1)
    else:
        d = datetime.fromisoformat(date)

    start_str = d.replace(hour=start_hour, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%S")
    end_str = d.replace(hour=end_hour, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%S")

    client = _get_client()
    suggestions = client.find_meeting_times(
        attendees=addr_list, start=start_str, end=end_str,
        duration_minutes=duration,
        timezone=cfg.get("timezone", "UTC"),
    )

    if _wants_json(as_json):
        click.echo(to_json_envelope(suggestions))
    else:
        if not suggestions:
            print_error("No available meeting slots found.")
        else:
            console.print(f"[bold cyan]Available slots ({len(suggestions)})[/bold cyan]")
            print_meeting_suggestions(suggestions)


@click.command("people-search")
@click.argument("query")
@click.option("--max", "-n", "max_count", default=10, type=int, help="Max results")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def people_search(query: str, max_count: int, as_json: bool):
    """Search people by name for attendee autocomplete."""
    client = _get_client()
    results = client.search_people(query, top=max_count)
    if _wants_json(as_json):
        click.echo(to_json_envelope(results))
    else:
        if not results:
            print_error("No people found.")
        else:
            print_people(results)



"""Structured exception hierarchy for outlook-cli."""

from __future__ import annotations


class OutlookCliError(Exception):
    """Base exception for all outlook-cli errors."""


class TokenExpiredError(OutlookCliError):
    """401 — bearer token expired or revoked."""


class RateLimitError(OutlookCliError):
    """429 — API rate limit hit after max retries."""


class ResourceNotFoundError(OutlookCliError):
    """Folder, calendar, category, signature, or message not found."""


class AuthRequiredError(OutlookCliError):
    """No token available — user must run 'outlook login'."""


def error_code_for_exception(exc: Exception) -> str:
    """Map exception to a structured error code string."""
    mapping = {
        TokenExpiredError: "session_expired",
        RateLimitError: "rate_limited",
        ResourceNotFoundError: "not_found",
        AuthRequiredError: "not_authenticated",
    }
    return mapping.get(type(exc), "unknown_error")



"""Shared helpers for all command modules."""

from __future__ import annotations

import sys

import click

from ..auth import get_token, login as do_login, verify_token
from ..client import OutlookClient
from ..config import load_config
from ..exceptions import AuthRequiredError, OutlookCliError, ResourceNotFoundError, TokenExpiredError, error_code_for_exception
from ..formatter import (
    console,
    print_attachments,
    print_calendars,
    print_categories,
    print_contacts,
    print_email,
    print_email_raw,
    print_error,
    print_event_detail,
    print_events,
    print_folders,
    print_inbox,
    print_meeting_suggestions,
    print_people,
    print_success,
    print_whoami,
)
from ..serialization import error_json, save_json, to_json, to_json_envelope

cfg = load_config()

# Cache client instance per session so auto-relogin can invalidate it
_client_cache: dict[str, OutlookClient] = {}


def _get_client() -> OutlookClient:
    if "c" not in _client_cache:
        try:
            token = get_token()
        except (AuthRequiredError, RuntimeError) as e:
            print_error(str(e))
            sys.exit(1)
        _client_cache["c"] = OutlookClient(token)
    return _client_cache["c"]


def _is_piped() -> bool:
    """True when stdout is not a terminal (piped to another command or file)."""
    return not sys.stdout.isatty()


def _wants_json(as_json: bool) -> bool:
    """True if JSON output is needed: explicit --json flag OR piped stdout."""
    return as_json or _is_piped()


def _is_json_mode() -> bool:
    """Check JSON mode from Click context (used by error handler)."""
    ctx = click.get_current_context(silent=True)
    explicit = bool(ctx and ctx.params.get("as_json"))
    return explicit or _is_piped()


def _handle_api_error(fn):
    """Decorator to catch common API errors. Auto re-login on 401."""
    import functools

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except TokenExpiredError:
            if _is_json_mode():
                click.echo(error_json("session_expired", "Token expired. Attempting re-login..."))
            else:
                print_error("Token expired. Attempting re-login...")
            try:
                token = do_login()
                print_success("Re-login successful. Retrying...")
                _client_cache.clear()
                return fn(*args, **kwargs)
            except Exception:
                if _is_json_mode():
                    click.echo(error_json("auth_failed", "Auto re-login failed. Run: outlook login --force"))
                else:
                    print_error("Auto re-login failed. Run: outlook login --force")
                sys.exit(1)
        except OutlookCliError as e:
            if _is_json_mode():
                click.echo(error_json(error_code_for_exception(e), str(e)))
            else:
                print_error(str(e))
            sys.exit(1)
        except Exception as e:
            if _is_json_mode():
                click.echo(error_json("unknown_error", str(e)))
            else:
                print_error(f"Error: {e}")
            sys.exit(1)

    return wrapper



"""Mail commands: inbox, read, thread, send, draft, draft-send, reply, reply-draft, forward."""

from __future__ import annotations

import click

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    cfg,
    console,
    print_email,
    print_email_raw,
    print_inbox,
    print_success,
    save_json,
    to_json_envelope,
)


@click.command()
@click.option("--max", "-n", "max_count", default=None, type=int, help="Number of messages")
@click.option("--unread", is_flag=True, help="Show only unread messages")
@click.option("--from", "from_filter", default=None, help="Filter by sender (name or email)")
@click.option("--subject", default=None, help="Filter by subject")
@click.option("--after", default=None, help="After date (YYYY-MM-DD)")
@click.option("--before", default=None, help="Before date (YYYY-MM-DD)")
@click.option("--has-attachments", is_flag=True, help="Only messages with attachments")
@click.option("--category", default=None, help="Filter by category name")
@click.option("--no-category", "no_category", is_flag=True, help="Only uncategorized messages")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--output", "-o", type=click.Path(), help="Save output to file")
@_handle_api_error
def inbox(
    max_count: int | None,
    unread: bool,
    from_filter: str | None,
    subject: str | None,
    after: str | None,
    before: str | None,
    has_attachments: bool,
    category: str | None,
    no_category: bool,
    as_json: bool,
    output: str | None,
):
    """Show inbox messages."""
    client = _get_client()
    top = max_count or cfg["max_messages"]
    has_filters = any([unread, from_filter, subject, after, before, has_attachments, category, no_category])

    messages = client.get_messages(
        folder="Inbox",
        top=top,
        unread_only=unread,
        filter_from=from_filter,
        filter_subject=subject,
        filter_after=after,
        filter_before=before,
        filter_has_attachments=has_attachments,
        filter_category=category,
        filter_no_category=no_category,
    )

    if _wants_json(as_json):
        if output:
            save_json(messages, output)
            print_success(f"Saved to {output}")
        else:
            click.echo(to_json_envelope(messages))
    else:
        # Show folder summary header
        if not has_filters:
            try:
                folder_info = client.get_folder("Inbox")
                console.print(
                    f"[bold cyan]Inbox[/bold cyan]  "
                    f"[dim]{folder_info.unread_count} unread / {folder_info.total_count} total[/dim]"
                )
            except Exception:
                pass
        if not messages:
            print_success("No messages found.")
        else:
            print_inbox(messages)


@click.command()
@click.argument("message_id")
@click.option("--raw", is_flag=True, help="Show raw HTML body")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def read(message_id: str, raw: bool, as_json: bool):
    """Read an email by its display number."""
    client = _get_client()
    email = client.get_message(message_id)

    if _wants_json(as_json):
        click.echo(to_json_envelope(email))
    elif raw:
        print_email_raw(email)
    else:
        print_email(email)

    # Auto mark as read
    if not email.is_read:
        try:
            client.mark_read(message_id)
        except Exception:
            pass


@click.command()
@click.argument("message_id")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def thread(message_id: str, as_json: bool):
    """Show the full conversation thread for a message."""
    from ..formatter import print_thread

    client = _get_client()
    messages = client.get_thread(message_id)

    if _wants_json(as_json):
        click.echo(to_json_envelope(messages))
    else:
        if len(messages) <= 1:
            print_success("This message is not part of a conversation thread.")
            if messages:
                print_email(messages[0])
        else:
            print_thread(messages)


@click.command()
@click.argument("to")
@click.argument("subject")
@click.argument("body")
@click.option("--cc", multiple=True, help="CC recipients")
@click.option("--html", "is_html", is_flag=True, help="Send body as HTML")
@click.option("--signature", "-s", "sig_name", default=None, help="Append a saved signature")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--yes", "-y", is_flag=True, help="Skip send confirmation")
@_handle_api_error
def send(to: str, subject: str, body: str, cc: tuple, is_html: bool, sig_name: str | None, as_json: bool, yes: bool):
    """Send an email. TO can be comma-separated for multiple recipients."""
    from ..signature_manager import append_signature, get_signature

    sig_name = sig_name or cfg.get("default_signature")
    if sig_name:
        sig_html = get_signature(sig_name)
        body, is_html = append_signature(body, sig_html, is_html)

    to_list = [addr.strip() for addr in to.split(",")]
    cc_list = list(cc) if cc else None

    if not yes:
        console.print(f"  [bold]To:[/bold] {', '.join(to_list)}")
        if cc_list:
            console.print(f"  [bold]CC:[/bold] {', '.join(cc_list)}")
        console.print(f"  [bold]Subject:[/bold] {subject}")
        console.print(f"  [bold]Body:[/bold] {body[:100]}{'...' if len(body) > 100 else ''}")
        click.confirm("Send this email?", abort=True)

    client = _get_client()
    client.send_mail(to=to_list, subject=subject, body=body, cc=cc_list, html=is_html)

    if _wants_json(as_json):
        click.echo(to_json_envelope({"status": "sent", "to": to_list, "subject": subject}))
    else:
        print_success(f"Mail sent to {to}")


@click.command()
@click.argument("to")
@click.argument("subject")
@click.argument("body")
@click.option("--cc", multiple=True, help="CC recipients")
@click.option("--html", "is_html", is_flag=True, help="Send body as HTML")
@click.option("--signature", "-s", "sig_name", default=None, help="Append a saved signature")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def draft(to: str, subject: str, body: str, cc: tuple, is_html: bool, sig_name: str | None, as_json: bool):
    """Create a draft email without sending. TO can be comma-separated."""
    from ..signature_manager import append_signature, get_signature

    sig_name = sig_name or cfg.get("default_signature")
    if sig_name:
        sig_html = get_signature(sig_name)
        body, is_html = append_signature(body, sig_html, is_html)

    client = _get_client()
    to_list = [addr.strip() for addr in to.split(",")]
    cc_list = list(cc) if cc else None
    email = client.create_draft(to=to_list, subject=subject, body=body, cc=cc_list, html=is_html)

    if _wants_json(as_json):
        click.echo(to_json_envelope(email))
    else:
        print_success(f"Draft created: {subject} (to: {to})")


@click.command(name="draft-send")
@click.argument("message_id")
@click.option("--yes", "-y", is_flag=True, help="Skip send confirmation")
@_handle_api_error
def draft_send(message_id: str, yes: bool):
    """Send an existing draft by its message number."""
    client = _get_client()
    if not yes:
        email = client.get_message(message_id)
        console.print(f"  [bold]To:[/bold] {', '.join(r.address for r in email.to)}")
        if email.cc:
            console.print(f"  [bold]CC:[/bold] {', '.join(r.address for r in email.cc)}")
        console.print(f"  [bold]Subject:[/bold] {email.subject}")
        click.confirm(f"Send draft #{message_id}?", abort=True)
    client.send_draft(message_id)
    print_success(f"Draft #{message_id} sent")


@click.command()
@click.argument("message_id")
@click.argument("body")
@click.option("--all", "reply_all", is_flag=True, help="Reply to all recipients")
@click.option("--yes", "-y", is_flag=True, help="Skip send confirmation")
@_handle_api_error
def reply(message_id: str, body: str, reply_all: bool, yes: bool):
    """Reply to an email."""
    client = _get_client()
    if not yes:
        action = "Reply all" if reply_all else "Reply"
        console.print(f"  [bold]{action} to #{message_id}[/bold]")
        console.print(f"  [bold]Body:[/bold] {body[:100]}{'...' if len(body) > 100 else ''}")
        click.confirm("Send this reply?", abort=True)
    client.reply(message_id, body, reply_all=reply_all)
    action = "Reply all" if reply_all else "Reply"
    print_success(f"{action} sent for message #{message_id}")


@click.command(name="reply-draft")
@click.argument("message_id")
@click.argument("body", default="")
@click.option("--all", "reply_all", is_flag=True, help="Reply to all recipients")
@click.option("--html", "is_html", is_flag=True, help="Body is HTML")
@click.option("--signature", "-s", "sig_name", default=None, help="Append a saved signature")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def reply_draft(message_id: str, body: str, reply_all: bool, is_html: bool, sig_name: str | None, as_json: bool):
    """Create a reply draft without sending."""
    from ..signature_manager import append_signature, get_signature

    sig_name = sig_name or cfg.get("default_signature")
    if sig_name and body:
        sig_html = get_signature(sig_name)
        body, is_html = append_signature(body, sig_html, is_html)

    client = _get_client()
    email = client.create_reply_draft(message_id, comment=body, reply_all=reply_all, html=is_html)
    action = "Reply-all" if reply_all else "Reply"
    if _wants_json(as_json):
        click.echo(to_json_envelope(email))
    else:
        print_success(f"{action} draft created for message #{message_id}")


@click.command()
@click.argument("message_id")
@click.argument("to")
@click.option("--comment", "-c", default="", help="Add a comment to the forwarded message")
@click.option("--yes", "-y", is_flag=True, help="Skip send confirmation")
@_handle_api_error
def forward(message_id: str, to: str, comment: str, yes: bool):
    """Forward an email."""
    to_list = [addr.strip() for addr in to.split(",")]
    if not yes:
        console.print(f"  [bold]Forward #{message_id} to:[/bold] {', '.join(to_list)}")
        if comment:
            console.print(f"  [bold]Comment:[/bold] {comment[:100]}{'...' if len(comment) > 100 else ''}")
        click.confirm("Forward this email?", abort=True)
    client = _get_client()
    client.forward(message_id, to_list, comment=comment)
    print_success(f"Message #{message_id} forwarded to {to}")



"""Attachment commands: attachments."""

from __future__ import annotations

import base64
from pathlib import Path

import click

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    print_attachments,
    print_success,
    to_json_envelope,
)


@click.command()
@click.argument("message_id")
@click.option("-d", "--download", is_flag=True, help="Download all attachments")
@click.option("--save-to", type=click.Path(), default=".", help="Download directory")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def attachments(message_id: str, download: bool, save_to: str, as_json: bool):
    """List or download attachments for a message."""
    client = _get_client()
    atts = client.get_attachments(message_id)

    if not atts:
        print_success("No attachments.")
        return

    if _wants_json(as_json):
        click.echo(to_json_envelope(atts))
        return

    print_attachments(atts)

    if download:
        save_path = Path(save_to)
        save_path.mkdir(parents=True, exist_ok=True)
        for att in atts:
            if att.content_bytes:
                file_path = save_path / att.name
                file_path.write_bytes(base64.b64decode(att.content_bytes))
                print_success(f"  Saved: {file_path}")
            else:
                full = client.download_attachment(message_id, att.id)
                if full.content_bytes:
                    file_path = save_path / full.name
                    file_path.write_bytes(base64.b64decode(full.content_bytes))
                    print_success(f"  Saved: {file_path}")



"""Schedule commands: schedule, schedule-list, schedule-cancel, schedule-draft."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone

import click
from rich.table import Table

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    cfg,
    console,
    print_error,
    print_success,
    to_json_envelope,
)


def _parse_schedule_time(s: str) -> datetime:
    """Parse schedule time from various formats.

    Supports:
      +30m, +1h, +2h30m           relative offset
      tomorrow 09:00              relative day
      today 17:00                 relative day
      2024-03-15T10:00            ISO format
      2024-03-15 10:00            space-separated
    """
    import re
    from datetime import timezone as tz

    now = datetime.now(tz.utc)

    # Relative offset: +30m, +1h, +2h30m
    offset_match = re.match(r'^\+(?:(\d+)h)?(?:(\d+)m)?$', s)
    if offset_match:
        hours = int(offset_match.group(1) or 0)
        minutes = int(offset_match.group(2) or 0)
        if hours == 0 and minutes == 0:
            raise click.BadParameter(f"Invalid offset: {s}")
        return now + timedelta(hours=hours, minutes=minutes)

    # today/tomorrow HH:MM
    day_match = re.match(r'^(today|tomorrow)\s+(\d{1,2}:\d{2})$', s, re.IGNORECASE)
    if day_match:
        day_word, time_str = day_match.groups()
        local_now = datetime.now().astimezone()
        h, m = map(int, time_str.split(":"))
        target = local_now.replace(hour=h, minute=m, second=0, microsecond=0)
        if day_word.lower() == "tomorrow":
            target += timedelta(days=1)
        return target.astimezone(tz.utc)

    # ISO-like: 2024-03-15T10:00 or 2024-03-15 10:00
    try:
        s = s.replace(" ", "T", 1) if " " in s and "T" not in s else s
        dt = datetime.fromisoformat(s)
        if dt.tzinfo is None:
            dt = dt.astimezone()
        return dt.astimezone(tz.utc)
    except ValueError:
        pass

    raise click.BadParameter(
        f"Cannot parse '{s}'. Use: +30m, +1h, tomorrow 09:00, or 2024-03-15T10:00"
    )


def _print_schedule_entries(entries: list[dict]) -> None:
    table = Table(show_header=True, header_style="bold cyan", box=None, pad_edge=False)
    table.add_column("#", style="dim", width=4, justify="right")
    table.add_column("To", width=28, no_wrap=True)
    table.add_column("Subject", ratio=1, no_wrap=True, overflow="ellipsis")
    table.add_column("Scheduled", width=16, no_wrap=True, justify="right")
    table.add_column("", width=6)

    now = datetime.now(timezone.utc)
    for i, entry in enumerate(entries, 1):
        to_str = ", ".join(entry.get("to", []))
        sched = entry.get("scheduled_at", "")
        try:
            sched_dt = datetime.fromisoformat(sched.replace("Z", "+00:00"))
            local_dt = sched_dt.astimezone(datetime.now().astimezone().tzinfo)
            sched_display = local_dt.strftime("%Y-%m-%d %H:%M")
        except (ValueError, AttributeError):
            sched_display = sched
        has_draft = bool(entry.get("message_id"))
        src_tag = "[cyan]draft[/cyan]" if has_draft else "[dim]queued[/dim]"
        table.add_row(str(i), to_str[:28], entry.get("subject", "")[:50], sched_display, src_tag)

    console.print(table)


@click.command()
@click.argument("to")
@click.argument("subject")
@click.argument("body")
@click.argument("at")
@click.option("--cc", multiple=True, help="CC recipients")
@click.option("--html", "is_html", is_flag=True, help="Send body as HTML")
@click.option("--signature", "-s", "sig_name", default=None, help="Append a saved signature")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def schedule(to: str, subject: str, body: str, at: str, cc: tuple, is_html: bool, sig_name: str | None, as_json: bool, yes: bool):
    """Schedule an email to be sent later.

    AT is the scheduled send time. Accepts:
    +30m, +1h, +2h30m (relative), tomorrow 09:00, 2024-03-15T10:00
    """
    from ..signature_manager import append_signature, get_signature

    send_at = _parse_schedule_time(at)

    sig_name = sig_name or cfg.get("default_signature")
    if sig_name:
        sig_html = get_signature(sig_name)
        body, is_html = append_signature(body, sig_html, is_html)

    to_list = [addr.strip() for addr in to.split(",")]
    cc_list = list(cc) if cc else None

    if not yes:
        local_send = send_at.astimezone(datetime.now().astimezone().tzinfo)
        console.print(f"  [bold]To:[/bold] {', '.join(to_list)}")
        if cc_list:
            console.print(f"  [bold]CC:[/bold] {', '.join(cc_list)}")
        console.print(f"  [bold]Subject:[/bold] {subject}")
        console.print(f"  [bold]Body:[/bold] {body[:100]}{'...' if len(body) > 100 else ''}")
        console.print(f"  [bold]Scheduled:[/bold] {local_send.strftime('%Y-%m-%d %H:%M')}")
        click.confirm("Schedule this email?", abort=True)

    client = _get_client()
    client.schedule_send(
        to=to_list, subject=subject, body=body, cc=cc_list,
        html=is_html, send_at=send_at.strftime("%Y-%m-%dT%H:%M:%SZ"),
    )

    if _wants_json(as_json):
        click.echo(to_json_envelope({"status": "scheduled", "to": to_list, "subject": subject, "scheduled_at": send_at.isoformat()}))
    else:
        local_send = send_at.astimezone(datetime.now().astimezone().tzinfo)
        print_success(f"Email scheduled to {to} at {local_send.strftime('%Y-%m-%d %H:%M')}")


@click.command(name="schedule-list")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def schedule_list(as_json: bool):
    """List scheduled (pending) emails."""
    client = _get_client()
    entries = client.get_scheduled_list()

    if _wants_json(as_json):
        click.echo(to_json_envelope(entries))
    else:
        if not entries:
            print_success("No scheduled emails.")
        else:
            console.print("[bold cyan]Scheduled Emails[/bold cyan]")
            _print_schedule_entries(entries)


@click.command(name="schedule-cancel")
@click.argument("index", type=int)
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def schedule_cancel(index: int, yes: bool):
    """Cancel a scheduled email by its list number.

    For draft entries: deletes the draft from server (prevents sending).
    For queued entries: removes local tracking only.
    Run schedule-list to see numbers.
    """
    client = _get_client()
    entries = client.get_scheduled_list()
    if index < 1 or index > len(entries):
        print_error(f"Invalid index #{index}. Run 'outlook schedule-list' to see entries.")
        return

    entry = entries[index - 1]
    if not yes:
        console.print(f"  [bold]To:[/bold] {', '.join(entry['to'])}")
        console.print(f"  [bold]Subject:[/bold] {entry['subject']}")
        console.print(f"  [bold]Scheduled:[/bold] {entry['scheduled_at']}")
        click.confirm(f"Remove scheduled entry #{index}?", abort=True)

    result = client.cancel_scheduled_entry(index)
    if result and result.get("server_deleted"):
        print_success(f"Scheduled email #{index} cancelled and draft deleted: {entry['subject']}")
    else:
        print_success(f"Scheduled entry #{index} removed: {entry['subject']}")


@click.command(name="schedule-draft")
@click.argument("message_id")
@click.argument("at")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def schedule_draft(message_id: str, at: str, yes: bool):
    """Schedule an existing draft to be sent later.

    AT is the scheduled send time. Accepts:
    +30m, +1h, +2h30m (relative), tomorrow 09:00, 2024-03-15T10:00
    """
    send_at = _parse_schedule_time(at)
    client = _get_client()

    if not yes:
        email = client.get_message(message_id)
        local_send = send_at.astimezone(datetime.now().astimezone().tzinfo)
        console.print(f"  [bold]To:[/bold] {', '.join(r.address for r in email.to)}")
        if email.cc:
            console.print(f"  [bold]CC:[/bold] {', '.join(r.address for r in email.cc)}")
        console.print(f"  [bold]Subject:[/bold] {email.subject}")
        console.print(f"  [bold]Scheduled:[/bold] {local_send.strftime('%Y-%m-%d %H:%M')}")
        click.confirm(f"Schedule draft #{message_id}?", abort=True)

    client.schedule_draft(message_id, send_at.strftime("%Y-%m-%dT%H:%M:%SZ"))

    local_send = send_at.astimezone(datetime.now().astimezone().tzinfo)
    print_success(f"Draft #{message_id} scheduled for {local_send.strftime('%Y-%m-%d %H:%M')}")



"""Tests for thread feature — get_thread client method and print_thread formatter."""

from __future__ import annotations

from io import StringIO
from unittest.mock import MagicMock, patch

from outlook_cli.formatter import print_thread
from outlook_cli.models import Email, EmailAddress


def _make_email(subject: str, sender: str, num: int, is_read: bool = True) -> Email:
    """Helper to create a minimal Email for testing."""
    from datetime import datetime, timezone
    return Email(
        id=f"id_{num}",
        subject=subject,
        sender=EmailAddress(name=sender, address=f"{sender.lower().replace(' ', '.')}@contoso.com"),
        to=[EmailAddress(name="Bob", address="bob@contoso.com")],
        cc=[],
        received=datetime(2026, 3, 10, 9, 0, tzinfo=timezone.utc),
        preview=f"Preview of {subject}",
        body=f"Body of {subject}",
        body_type="Text",
        is_read=is_read,
        has_attachments=False,
        importance="Normal",
        conversation_id="conv_abc123",
        display_num=num,
    )


class TestPrintThread:
    def test_shows_all_messages(self, capsys):
        messages = [
            _make_email("Initial question", "Alice", 1),
            _make_email("Re: Initial question", "Bob", 2),
            _make_email("Re: Re: Initial question", "Alice", 3),
        ]
        print_thread(messages)
        captured = capsys.readouterr()
        assert "3 messages" in captured.err  # Rich goes to stderr
        assert "Alice" in captured.err
        assert "Bob" in captured.err

    def test_unread_marker(self, capsys):
        messages = [
            _make_email("Hello", "Alice", 1, is_read=True),
            _make_email("Re: Hello", "Bob", 2, is_read=False),
        ]
        print_thread(messages)
        captured = capsys.readouterr()
        assert "*" in captured.err  # unread marker for Bob's message

    def test_single_message_thread(self):
        """A single-message thread should still work."""
        messages = [_make_email("Solo message", "Alice", 1)]
        # Should not raise
        print_thread(messages)

    def test_long_body_truncation(self, capsys):
        """Messages with >20 lines should be truncated."""
        email = _make_email("Long email", "Alice", 1)
        email.body = "\n".join(f"Line {i}" for i in range(50))
        print_thread([email])
        captured = capsys.readouterr()
        assert "more lines" in captured.err


class TestGetThread:
    def test_returns_conversation_messages(self):
        """get_thread should search by subject and filter by ConversationId client-side."""
        from outlook_cli.client import OutlookClient

        with patch.object(OutlookClient, '__init__', lambda self, token: None):
            client = OutlookClient.__new__(OutlookClient)
            client._id_map = {"1": "real_id_1"}
            client._next_num = 10

            mock_email = _make_email("Re: Test", "Alice", 1)
            client.get_message = MagicMock(return_value=mock_email)

            # API returns messages from search — includes same + different conversations
            api_response = {
                "value": [
                    {
                        "Id": "msg_1",
                        "Subject": "Test",
                        "ConversationId": "conv_abc123",
                        "ReceivedDateTime": "2026-03-10T09:00:00Z",
                    },
                    {
                        "Id": "msg_2",
                        "Subject": "Re: Test",
                        "ConversationId": "conv_abc123",
                        "ReceivedDateTime": "2026-03-10T10:00:00Z",
                    },
                    {
                        "Id": "msg_other",
                        "Subject": "Test something else",
                        "ConversationId": "conv_DIFFERENT",
                        "ReceivedDateTime": "2026-03-10T11:00:00Z",
                    },
                ]
            }
            client._get = MagicMock(return_value=api_response)
            client._assign_display_nums = MagicMock()

            messages = client.get_thread("1")

            # Should only include messages with matching ConversationId
            assert len(messages) == 2
            assert messages[0].subject == "Test"
            assert messages[1].subject == "Re: Test"

            # Verify $search was used (not $filter)
            call_args = client._get.call_args
            assert "$search" in call_args[1]["params"]

    def test_no_conversation_id_returns_single(self):
        """If message has no conversation_id, return just that message."""
        from outlook_cli.client import OutlookClient

        with patch.object(OutlookClient, '__init__', lambda self, token: None):
            client = OutlookClient.__new__(OutlookClient)
            client._id_map = {"1": "real_id_1"}
            client._next_num = 10

            mock_email = _make_email("Solo", "Alice", 1)
            mock_email.conversation_id = ""
            client.get_message = MagicMock(return_value=mock_email)

            messages = client.get_thread("1")
            assert len(messages) == 1
            assert messages[0].subject == "Solo"



from __future__ import annotations

from pathlib import Path

import yaml

from .constants import CONFIG_FILE

DEFAULTS = {
    "max_messages": 25,
    "default_folder": "Inbox",
    "default_signature": None,
    "timezone": "UTC",
    "browser": {
        "headless": False,
        "timeout": 120,
    },
    "output_format": "table",
}


def load_config(path: Path | None = None) -> dict:
    """Load YAML config, falling back to defaults for missing keys."""
    cfg = dict(DEFAULTS)
    p = path or CONFIG_FILE
    if p.exists():
        with open(p) as f:
            user = yaml.safe_load(f) or {}
        _deep_merge(cfg, user)
    return cfg


def _deep_merge(base: dict, override: dict) -> None:
    for k, v in override.items():
        if isinstance(v, dict) and isinstance(base.get(k), dict):
            _deep_merge(base[k], v)
        else:
            base[k] = v



"""Category commands: categories, categorize, uncategorize, category-create/rename/clear/delete."""

from __future__ import annotations

import click

from ._common import (
    _get_client,
    _handle_api_error,
    _wants_json,
    console,
    get_token,
    print_categories,
    print_success,
    to_json_envelope,
)


@click.command()
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@_handle_api_error
def categories(as_json: bool):
    """List master categories with unread/total counts."""
    client = _get_client()
    resp = client.get_master_categories()
    cat_list = resp.get("Body", {}).get("CategoryDetailsList", [])

    if _wants_json(as_json):
        click.echo(to_json_envelope(cat_list))
    else:
        if not cat_list:
            print_success("No categories defined.")
        else:
            print_categories(cat_list)


@click.command()
@click.argument("message_ids", nargs=-1, required=True)
@click.argument("category")
@_handle_api_error
def categorize(message_ids: tuple, category: str):
    """Add a category to messages. Accepts multiple IDs."""
    client = _get_client()
    for mid in message_ids:
        result = client.add_category(mid, category)
        print_success(f"Message #{mid} categorized as: {', '.join(result)}")


@click.command()
@click.argument("message_ids", nargs=-1, required=True)
@click.argument("category")
@_handle_api_error
def uncategorize(message_ids: tuple, category: str):
    """Remove a category from messages. Accepts multiple IDs."""
    client = _get_client()
    for mid in message_ids:
        result = client.remove_category(mid, category)
        if result:
            print_success(f"Message #{mid} categories: {', '.join(result)}")
        else:
            print_success(f"Message #{mid} has no categories.")


@click.command("category-rename")
@click.argument("old_name")
@click.argument("new_name")
@click.option("--no-propagate", is_flag=True, help="Only rename master category, skip updating messages")
@_handle_api_error
def category_rename(old_name: str, new_name: str, no_propagate: bool):
    """Rename a master category and update all messages."""
    from ..category_manager import rename_category

    def on_progress(done, _total):
        console.print(f"  [dim]{done} messages updated...[/dim]")

    token = get_token()
    count = rename_category(token, old_name, new_name, propagate=not no_propagate, on_progress=on_progress)
    print_success(f"Renamed '{old_name}' → '{new_name}'")
    if count:
        print_success(f"  {count} messages updated")


@click.command("category-clear")
@click.argument("name")
@click.option("--folder", default=None, help="Limit to a specific folder")
@click.option("--max", "-n", "max_messages", type=int, default=None, help="Max messages to clear")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def category_clear(name: str, folder: str | None, max_messages: int | None, yes: bool):
    """Remove a category label from messages (does not delete master category)."""
    from ..category_manager import clear_category

    scope = f"in '{folder}'" if folder else "in all folders"
    limit = f" (max {max_messages})" if max_messages else ""
    if not yes:
        click.confirm(f"Remove '{name}' from messages {scope}{limit}?", abort=True)

    def on_progress(done, _total):
        console.print(f"  [dim]{done} messages cleared...[/dim]")

    token = get_token()
    count = clear_category(token, name, folder=folder, max_messages=max_messages, on_progress=on_progress)
    print_success(f"Cleared '{name}' from {count} messages")


@click.command("category-delete")
@click.argument("name")
@click.option("--no-propagate", is_flag=True, help="Only delete master category, skip clearing messages")
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation")
@_handle_api_error
def category_delete(name: str, no_propagate: bool, yes: bool):
    """Delete a master category and remove it from all messages."""
    from ..category_manager import clear_category, delete_category

    if not yes:
        click.confirm(f"Delete category '{name}' and remove from all messages?", abort=True)

    token = get_token()

    if not no_propagate:
        def on_progress(done, _total):
            console.print(f"  [dim]{done} messages cleared...[/dim]")

        count = clear_category(token, name, on_progress=on_progress)
        if count:
            print_success(f"  Cleared from {count} messages")

    delete_category(token, name)
    print_success(f"Deleted category '{name}'")


@click.command("category-create")
@click.argument("name")
@click.option("--color", type=int, default=15, help="Color index (0-24)")
@_handle_api_error
def category_create(name: str, color: int):
    """Create a new master category."""
    from ..category_manager import create_category
    token = get_token()
    create_category(token, name, color=color)
    print_success(f"Created category '{name}'")



from __future__ import annotations

import json
import time
from pathlib import Path

import httpx

from urllib.parse import quote

from .constants import BASE_URL, CACHE_DIR, DEFERRED_SEND_PROPERTY_ID, ID_MAP_FILE, OWA_SERVICE_URL, SCHEDULED_FILE, USER_AGENT
from .exceptions import RateLimitError, ResourceNotFoundError, TokenExpiredError
from .models import Attachment, Contact, Email, Event, Folder


def _build_query_params(
    unread_only: bool = False,
    filter_from: str | None = None,
    filter_subject: str | None = None,
    filter_after: str | None = None,
    filter_before: str | None = None,
    filter_has_attachments: bool = False,
    filter_category: str | None = None,
) -> tuple[str, str, bool]:
    """Build $filter and $search params.

    REST v2 limitations:
    - $filter and $search can't be combined
    - $filter doesn't support contains() on From
    - $search KQL supports from:, subject:, hasattachments:, received:

    Strategy: if text filters (from/subject) are used, build a KQL $search.
    Otherwise use $filter for IsRead/date (which supports $orderby).

    Returns (filter_str, search_str, needs_search).
    When needs_search is True, $orderby must be omitted.
    """
    has_text_filters = any([filter_from, filter_subject, filter_has_attachments])

    if has_text_filters:
        # Use $search with KQL — can't combine with $filter
        kql_parts: list[str] = []
        if filter_from:
            kql_parts.append(f"from:{filter_from}")
        if filter_subject:
            kql_parts.append(f"subject:{filter_subject}")
        if filter_has_attachments:
            kql_parts.append("hasattachments:true")
        if unread_only:
            kql_parts.append("isread:false")
        if filter_after:
            kql_parts.append(f"received>={filter_after}")
        if filter_before:
            kql_parts.append(f"received<={filter_before}")
        if filter_category:
            kql_parts.append(f'category:"{filter_category}"')
        return "", f'"{" ".join(kql_parts)}"', True

    # Pure $filter — supports $orderby
    filter_parts: list[str] = []
    if unread_only:
        filter_parts.append("IsRead eq false")
    if filter_after:
        filter_parts.append(f"ReceivedDateTime ge {filter_after}T00:00:00Z")
    if filter_before:
        filter_parts.append(f"ReceivedDateTime lt {filter_before}T23:59:59Z")
    if filter_category:
        filter_parts.append(f"Categories/any(c:c eq '{filter_category}')")
    return " and ".join(filter_parts), "", False


class OutlookClient:
    """HTTP client for Outlook REST API v2."""

    MAX_ID_MAP_SIZE = 500

    def __init__(self, token: str):
        self._token = token
        self._client = httpx.Client(
            base_url=BASE_URL,
            headers={
                "Authorization": f"Bearer {token}",
                "User-Agent": USER_AGENT,
                "Content-Type": "application/json",
            },
            timeout=30,
        )
        self._id_map: dict[str, str] = self._load_id_map()
        self._next_num: int = max((int(k) for k in self._id_map if k.isdigit()), default=0) + 1

    # ------------------------------------------------------------------
    # Mail
    # ------------------------------------------------------------------

    def get_messages(
        self,
        folder: str = "Inbox",
        top: int = 25,
        skip: int = 0,
        unread_only: bool = False,
        filter_from: str | None = None,
        filter_subject: str | None = None,
        filter_after: str | None = None,
        filter_before: str | None = None,
        filter_has_attachments: bool = False,
        filter_category: str | None = None,
        filter_no_category: bool = False,
        select: str | None = None,
    ) -> list[Email]:
        filter_str, search_str, needs_search = _build_query_params(
            unread_only=unread_only,
            filter_from=filter_from,
            filter_subject=filter_subject,
            filter_after=filter_after,
            filter_before=filter_before,
            filter_has_attachments=filter_has_attachments,
            filter_category=filter_category,
        )

        if not filter_no_category:
            # Standard fetch — server-side filtering is sufficient
            if needs_search:
                params: dict = {"$top": top, "$search": search_str}
                if select:
                    params["$select"] = select
                resp = self._get("/messages", params=params)
            else:
                params = {
                    "$top": top,
                    "$skip": skip,
                    "$orderby": "ReceivedDateTime desc",
                }
                if filter_str:
                    params["$filter"] = filter_str
                if select:
                    params["$select"] = select
                resp = self._get(f"/MailFolders/{folder}/messages", params=params)
            messages = [Email.from_api(m) for m in resp.get("value", [])]
        else:
            # Client-side filtering: over-fetch in pages until we have enough
            messages: list[Email] = []
            batch_size = top * 3  # fetch 3x to compensate for filtering
            current_skip = skip
            max_pages = 5  # safety limit
            for _ in range(max_pages):
                if needs_search:
                    params = {"$top": batch_size, "$search": search_str}
                    if select:
                        params["$select"] = select
                    resp = self._get("/messages", params=params)
                else:
                    params = {
                        "$top": batch_size,
                        "$skip": current_skip,
                        "$orderby": "ReceivedDateTime desc",
                    }
                    if filter_str:
                        params["$filter"] = filter_str
                    if select:
                        params["$select"] = select
                    resp = self._get(f"/MailFolders/{folder}/messages", params=params)
                batch = resp.get("value", [])
                if not batch:
                    break
                for m in batch:
                    email = Email.from_api(m)
                    if not email.categories:
                        messages.append(email)
                        if len(messages) >= top:
                            break
                if len(messages) >= top or len(batch) < batch_size:
                    break
                current_skip += batch_size
            messages = messages[:top]

        self._assign_display_nums(messages)
        return messages

    def get_message(self, message_id: str) -> Email:
        real_id = self._resolve_id(message_id)
        resp = self._get(f"/messages/{real_id}")
        email = Email.from_api(resp)
        email.display_num = int(message_id) if message_id.isdigit() else 0
        return email

    def get_thread(self, message_id: str, max_messages: int = 50) -> list[Email]:
        """Fetch all messages in the same conversation as the given message.

        REST v2 doesn't support $filter on ConversationId, so we search by
        the base subject (strip Re:/Fwd: prefixes) and then filter client-side
        by matching ConversationId. Results are sorted oldest-first.
        """
        import re

        email = self.get_message(message_id)
        conv_id = email.conversation_id
        if not conv_id:
            return [email]

        # Strip reply/forward prefixes to get the base subject for search
        base_subject = re.sub(
            r'^(Re|Fwd|İlt|Ynt|Fw|AW|SV|VS)\s*:\s*',
            '', email.subject, flags=re.IGNORECASE,
        ).strip()

        if not base_subject:
            return [email]

        # Search by subject, then filter by ConversationId client-side
        resp = self._get(
            "/messages",
            params={
                "$search": f'"subject:{base_subject}"',
                "$top": max_messages,
            },
        )
        all_msgs = [Email.from_api(m) for m in resp.get("value", [])]
        thread = [m for m in all_msgs if m.conversation_id == conv_id]

        # Sort chronologically (oldest first)
        thread.sort(key=lambda m: m.received)

        self._assign_display_nums(thread)
        return thread if thread else [email]

    def send_mail(
        self,
        to: list[str],
        subject: str,
        body: str,
        cc: list[str] | None = None,
        html: bool = False,
        send_at: str | None = None,
    ) -> None:
        message: dict = {
            "Subject": subject,
            "Body": {
                "ContentType": "HTML" if html else "Text",
                "Content": body,
            },
            "ToRecipients": [
                {"EmailAddress": {"Address": addr}} for addr in to
            ],
        }
        if cc:
            message["CcRecipients"] = [
                {"EmailAddress": {"Address": addr}} for addr in cc
            ]
        if send_at:
            message["SingleValueExtendedProperties"] = [{
                "PropertyId": DEFERRED_SEND_PROPERTY_ID,
                "Value": send_at,
            }]
        self._post("/sendmail", json={"Message": message})

    def create_draft(
        self,
        to: list[str],
        subject: str,
        body: str,
        cc: list[str] | None = None,
        html: bool = False,
    ) -> Email:
        payload: dict = {
            "Subject": subject,
            "Body": {
                "ContentType": "HTML" if html else "Text",
                "Content": body,
            },
            "ToRecipients": [
                {"EmailAddress": {"Address": addr}} for addr in to
            ],
        }
        if cc:
            payload["CcRecipients"] = [
                {"EmailAddress": {"Address": addr}} for addr in cc
            ]
        data = self._post("/messages", json=payload)
        return Email.from_api(data)

    def send_draft(self, message_id: str) -> None:
        real_id = self._resolve_id(message_id)
        self._post(f"/messages/{real_id}/send")

    def reply(self, message_id: str, comment: str, reply_all: bool = False) -> None:
        real_id = self._resolve_id(message_id)
        action = "replyall" if reply_all else "reply"
        self._post(f"/messages/{real_id}/{action}", json={"Comment": comment})

    def create_reply_draft(
        self,
        message_id: str,
        comment: str = "",
        reply_all: bool = False,
        html: bool = False,
    ) -> Email:
        real_id = self._resolve_id(message_id)
        action = "createreplyall" if reply_all else "createreply"

        if html and comment:
            # createReply only supports plain-text Comment.
            # Create empty draft first, then prepend HTML body before quoted reply.
            data = self._post(f"/messages/{real_id}/{action}", json={})
            draft_id = data["Id"]
            original_body = data.get("Body", {}).get("Content", "")
            # Insert user's HTML before the quoted original message
            if "<body>" in original_body:
                combined = original_body.replace("<body>", f"<body>{comment} ", 1)
            else:
                combined = comment + original_body
            data = self._patch(f"/messages/{draft_id}", json={
                "Body": {"ContentType": "HTML", "Content": combined},
            })
        else:
            payload = {}
            if comment:
                payload["Comment"] = comment
            data = self._post(f"/messages/{real_id}/{action}", json=payload)

        return Email.from_api(data)

    def forward(self, message_id: str, to: list[str], comment: str = "") -> None:
        real_id = self._resolve_id(message_id)
        payload = {
            "Comment": comment,
            "ToRecipients": [{"EmailAddress": {"Address": addr}} for addr in to],
        }
        self._post(f"/messages/{real_id}/forward", json=payload)

    def move_message(self, message_id: str, destination_folder: str) -> Email:
        real_id = self._resolve_id(message_id)
        folder_id = self._resolve_folder(destination_folder)
        resp = self._post(
            f"/messages/{real_id}/move",
            json={"DestinationId": folder_id},
        )
        return Email.from_api(resp)

    def _resolve_folder(self, name_or_id: str) -> str:
        """Resolve a folder display name to its ID. Pass through if already an ID."""
        if len(name_or_id) > 50:
            return name_or_id  # likely already an ID
        # Well-known folder names work directly with the API
        well_known = {
            "inbox", "drafts", "sentitems", "deleteditems",
            "junkemail", "archive", "outbox",
        }
        if name_or_id.lower() in well_known:
            return name_or_id
        # Search by display name
        folders = self.get_folders()
        for f in folders:
            if f.name.lower() == name_or_id.lower():
                return f.id
        raise ResourceNotFoundError(f"Folder '{name_or_id}' not found. Run 'outlook folders' to see available folders.")

    def delete_message(self, message_id: str) -> None:
        real_id = self._resolve_id(message_id)
        self._delete(f"/messages/{real_id}")

    def mark_read(self, message_id: str, is_read: bool = True) -> None:
        real_id = self._resolve_id(message_id)
        self._patch(f"/messages/{real_id}", json={"IsRead": is_read})

    # ------------------------------------------------------------------
    # Scheduled send
    # ------------------------------------------------------------------

    def schedule_send(
        self,
        to: list[str],
        subject: str,
        body: str,
        send_at: str,
        cc: list[str] | None = None,
        html: bool = False,
    ) -> dict:
        """Schedule an email via /sendmail with deferred send time.

        send_at must be ISO 8601 format (e.g. 2024-03-15T10:00:00Z).
        Returns the tracked schedule entry.
        """
        self.send_mail(to=to, subject=subject, body=body, cc=cc, html=html, send_at=send_at)
        entry = self._track_scheduled(
            to=to, cc=cc, subject=subject, send_at=send_at,
        )
        return entry

    def schedule_draft(self, message_id: str, send_at: str) -> dict:
        """Set deferred send time on an existing draft and send it."""
        real_id = self._resolve_id(message_id)
        # Read draft details for tracking
        msg = self._get(f"/messages/{real_id}", params={"$select": "Subject,ToRecipients,CcRecipients"})
        to = [r["EmailAddress"]["Address"] for r in msg.get("ToRecipients", [])]
        cc = [r["EmailAddress"]["Address"] for r in msg.get("CcRecipients", [])]
        subject = msg.get("Subject", "")

        resp = self._patch(f"/messages/{real_id}", json={
            "SingleValueExtendedProperties": [{
                "PropertyId": DEFERRED_SEND_PROPERTY_ID,
                "Value": send_at,
            }],
        })
        updated_id = resp.get("Id", real_id)
        self._post(f"/messages/{updated_id}/send")

        entry = self._track_scheduled(
            to=to, cc=cc or None, subject=subject, send_at=send_at,
            message_id=updated_id,
        )
        return entry

    def get_scheduled_list(self) -> list[dict]:
        """Get scheduled messages from local tracking + Drafts cross-check.

        REST v2 doesn't support $filter/$expand on extended properties,
        so we cross-reference local tracking with Drafts folder by subject
        to enrich entries with message_id for server-side cancellation.
        """
        entries = self._load_scheduled()
        if not entries:
            return entries

        # Cross-check with Drafts to find matching message IDs
        try:
            resp = self._get("/MailFolders/Drafts/messages", params={
                "$top": 50,
                "$select": "Id,Subject",
            })
            drafts_by_subject: dict[str, str] = {}
            for m in resp.get("value", []):
                drafts_by_subject[m.get("Subject", "")] = m["Id"]

            for entry in entries:
                if not entry.get("message_id"):
                    draft_id = drafts_by_subject.get(entry.get("subject", ""))
                    if draft_id:
                        entry["message_id"] = draft_id
        except Exception:
            pass  # server unavailable, show local-only

        return entries

    def cancel_scheduled_entry(self, index: int) -> dict | None:
        """Cancel a scheduled entry by its 1-based index.

        If a matching draft is found on server, deletes it.
        Always removes from local tracking.
        """
        # Use enriched list (with message_id from Drafts cross-check)
        enriched = self.get_scheduled_list()
        if index < 1 or index > len(enriched):
            return None

        removed = enriched[index - 1]

        # Remove from local tracking
        local = self._load_scheduled()
        if index - 1 < len(local):
            local.pop(index - 1)
            self._save_scheduled(local)

        # Try to delete the draft from server
        msg_id = removed.get("message_id")
        if msg_id:
            try:
                self._delete(f"/messages/{msg_id}")
                removed["server_deleted"] = True
            except Exception:
                removed["server_deleted"] = False

        return removed

    # ------------------------------------------------------------------
    # Scheduled tracking helpers
    # ------------------------------------------------------------------

    def _track_scheduled(
        self,
        to: list[str],
        subject: str,
        send_at: str,
        cc: list[str] | None = None,
        message_id: str | None = None,
    ) -> dict:
        from datetime import datetime as dt, timezone as tz
        entry = {
            "to": to,
            "cc": cc or [],
            "subject": subject,
            "scheduled_at": send_at,
            "created_at": dt.now(tz.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
        }
        if message_id:
            entry["message_id"] = message_id
        entries = self._load_scheduled()
        entries.append(entry)
        self._save_scheduled(entries)
        return entry

    def _load_scheduled(self) -> list[dict]:
        if SCHEDULED_FILE.exists():
            try:
                return json.loads(SCHEDULED_FILE.read_text())
            except (json.JSONDecodeError, OSError):
                pass
        return []

    def _save_scheduled(self, entries: list[dict]) -> None:
        CACHE_DIR.mkdir(parents=True, exist_ok=True)
        SCHEDULED_FILE.write_text(json.dumps(entries, indent=2))

    def search_messages(self, query: str, top: int = 25) -> list[Email]:
        params = {
            "$search": f'"{query}"',
            "$top": top,
        }
        resp = self._get("/messages", params=params)
        messages = [Email.from_api(m) for m in resp.get("value", [])]
        self._assign_display_nums(messages)
        return messages

    # ------------------------------------------------------------------
    # Folders
    # ------------------------------------------------------------------

    def get_folders(self) -> list[Folder]:
        resp = self._get("/MailFolders", params={"$top": 100})
        return [Folder.from_api(f) for f in resp.get("value", [])]

    def get_folder(self, folder_id: str) -> Folder:
        resp = self._get(f"/MailFolders/{folder_id}")
        return Folder.from_api(resp)

    # ------------------------------------------------------------------
    # Attachments
    # ------------------------------------------------------------------

    def get_attachments(self, message_id: str) -> list[Attachment]:
        real_id = self._resolve_id(message_id)
        resp = self._get(f"/messages/{real_id}/attachments")
        return [Attachment.from_api(a) for a in resp.get("value", [])]

    def download_attachment(self, message_id: str, attachment_id: str) -> Attachment:
        real_id = self._resolve_id(message_id)
        resp = self._get(f"/messages/{real_id}/attachments/{attachment_id}")
        return Attachment.from_api(resp)

    # ------------------------------------------------------------------
    # Calendar
    # ------------------------------------------------------------------

    def get_calendar_view(self, start: str, end: str, top: int = 50, calendar_name: str | None = None) -> list[Event]:
        params = {
            "startDateTime": start,
            "endDateTime": end,
            "$top": top,
            "$orderby": "Start/DateTime asc",
        }
        if calendar_name:
            cal_id = self._resolve_calendar(calendar_name)
            path = f"/calendars/{cal_id}/calendarview"
        else:
            path = "/calendarview"
        resp = self._get(path, params=params)
        events = [Event.from_api(e) for e in resp.get("value", [])]
        self._assign_event_display_nums(events)
        return events

    def get_events(self, top: int = 25) -> list[Event]:
        resp = self._get("/events", params={"$top": top, "$orderby": "Start/DateTime desc"})
        events = [Event.from_api(e) for e in resp.get("value", [])]
        self._assign_event_display_nums(events)
        return events

    def get_event(self, event_id: str) -> Event:
        real_id = self._resolve_id(event_id)
        resp = self._get(f"/events/{real_id}")
        event = Event.from_api(resp)
        event.display_num = int(event_id) if event_id.isdigit() else 0
        return event

    def create_event(
        self,
        subject: str,
        start: str,
        end: str,
        timezone: str = "UTC",
        attendees: list[str] | None = None,
        location: str | None = None,
        body: str | None = None,
        html: bool = False,
        is_all_day: bool = False,
        reminder_minutes: int | None = 15,
        is_online_meeting: bool = False,
        recurrence: dict | None = None,
    ) -> Event:
        payload: dict = {
            "Subject": subject,
            "Start": {"DateTime": start, "TimeZone": timezone},
            "End": {"DateTime": end, "TimeZone": timezone},
            "IsAllDay": is_all_day,
        }
        if attendees:
            payload["Attendees"] = [
                {"EmailAddress": {"Address": addr}, "Type": "Required"}
                for addr in attendees
            ]
        if location:
            payload["Location"] = {"DisplayName": location}
        if body:
            payload["Body"] = {
                "ContentType": "HTML" if html else "Text",
                "Content": body,
            }
        if reminder_minutes is not None:
            payload["IsReminderOn"] = True
            payload["ReminderMinutesBeforeStart"] = reminder_minutes
        if is_online_meeting:
            payload["IsOnlineMeeting"] = True
            payload["OnlineMeetingProvider"] = "TeamsForBusiness"
        if recurrence:
            payload["Recurrence"] = recurrence
        data = self._post("/events", json=payload)
        return Event.from_api(data)

    def get_event_instances(self, event_id: str, start: str, end: str, top: int = 50) -> list[Event]:
        """Get occurrences of a recurring event.

        If given an occurrence ID, resolves to its series master first.
        """
        real_id = self._resolve_id(event_id)
        # Check if this is an occurrence — need series master for /instances
        ev = self._get(f"/events/{real_id}", params={"$select": "Type,SeriesMasterId"})
        master_id = ev.get("SeriesMasterId") or real_id
        resp = self._get(f"/events/{master_id}/instances", params={
            "startDateTime": start,
            "endDateTime": end,
            "$top": top,
        })
        events = [Event.from_api(e) for e in resp.get("value", [])]
        self._assign_event_display_nums(events)
        return events

    def update_event(self, event_id: str, **kwargs) -> Event:
        """Update event fields. Accepts: subject, start, end, timezone,
        location, body, html, is_all_day, attendees (full replacement)."""
        real_id = self._resolve_id(event_id)
        payload: dict = {}
        tz = kwargs.get("timezone", "UTC")
        if "subject" in kwargs:
            payload["Subject"] = kwargs["subject"]
        if "start" in kwargs:
            payload["Start"] = {"DateTime": kwargs["start"], "TimeZone": tz}
        if "end" in kwargs:
            payload["End"] = {"DateTime": kwargs["end"], "TimeZone": tz}
        if "location" in kwargs:
            payload["Location"] = {"DisplayName": kwargs["location"]}
        if "body" in kwargs:
            payload["Body"] = {
                "ContentType": "HTML" if kwargs.get("html") else "Text",
                "Content": kwargs["body"],
            }
        if "is_all_day" in kwargs:
            payload["IsAllDay"] = kwargs["is_all_day"]
        if "attendees" in kwargs:
            payload["Attendees"] = [
                {"EmailAddress": {"Address": addr}, "Type": "Required"}
                for addr in kwargs["attendees"]
            ]
        data = self._patch(f"/events/{real_id}", json=payload)
        return Event.from_api(data)

    def add_event_attendees(self, event_id: str, new_addrs: list[str]) -> Event:
        """Add attendees to an existing event without removing current ones."""
        real_id = self._resolve_id(event_id)
        current = self._get(f"/events/{real_id}", params={"$select": "Attendees"})
        existing = current.get("Attendees", [])
        existing_addrs = {a["EmailAddress"]["Address"].lower() for a in existing}
        for addr in new_addrs:
            if addr.lower() not in existing_addrs:
                existing.append({"EmailAddress": {"Address": addr}, "Type": "Required"})
        data = self._patch(f"/events/{real_id}", json={"Attendees": existing})
        return Event.from_api(data)

    def remove_event_attendees(self, event_id: str, remove_addrs: list[str]) -> Event:
        """Remove attendees from an existing event."""
        real_id = self._resolve_id(event_id)
        current = self._get(f"/events/{real_id}", params={"$select": "Attendees"})
        existing = current.get("Attendees", [])
        remove_lower = {a.lower() for a in remove_addrs}
        filtered = [a for a in existing if a["EmailAddress"]["Address"].lower() not in remove_lower]
        data = self._patch(f"/events/{real_id}", json={"Attendees": filtered})
        return Event.from_api(data)

    def delete_event(self, event_id: str) -> None:
        real_id = self._resolve_id(event_id)
        self._delete(f"/events/{real_id}")

    def respond_to_event(self, event_id: str, response: str, comment: str = "", send_response: bool = True) -> None:
        """Respond to a meeting. response: accept, decline, tentativelyaccept."""
        real_id = self._resolve_id(event_id)
        payload = {"SendResponse": send_response}
        if comment:
            payload["Comment"] = comment
        self._post(f"/events/{real_id}/{response}", json=payload)

    def find_meeting_times(
        self,
        attendees: list[str],
        start: str,
        end: str,
        duration_minutes: int = 60,
        timezone: str = "UTC",
        max_candidates: int = 5,
    ) -> list[dict]:
        payload = {
            "Attendees": [
                {"Type": "Required", "EmailAddress": {"Address": addr}}
                for addr in attendees
            ],
            "TimeConstraint": {
                "Timeslots": [{
                    "Start": {"DateTime": start, "TimeZone": timezone},
                    "End": {"DateTime": end, "TimeZone": timezone},
                }]
            },
            "MeetingDuration": f"PT{duration_minutes}M",
            "MaxCandidates": max_candidates,
        }
        resp = self._post("/findMeetingTimes", json=payload)
        return resp.get("MeetingTimeSuggestions", [])

    def search_people(self, query: str, top: int = 10) -> list[dict]:
        resp = self._get("/people", params={"$search": query, "$top": top})
        return resp.get("value", [])

    def get_calendars(self) -> list[dict]:
        resp = self._get("/calendars", params={"$top": 50})
        return resp.get("value", [])

    def _resolve_calendar(self, name: str) -> str:
        """Resolve a calendar display name to its ID."""
        cals = self.get_calendars()
        # Exact match first
        for c in cals:
            if c.get("Name", "").lower() == name.lower():
                return c["Id"]
        # Partial match
        for c in cals:
            if name.lower() in c.get("Name", "").lower():
                return c["Id"]
        available = ", ".join(c.get("Name", "") for c in cals)
        raise ResourceNotFoundError(f"Calendar '{name}' not found. Available: {available}")

    def _assign_event_display_nums(self, events: list[Event]) -> None:
        """Assign display numbers to events using the shared ID map."""
        for ev in events:
            existing = next(
                (k for k, v in self._id_map.items() if v == ev.id and k.isdigit()),
                None,
            )
            if existing:
                ev.display_num = int(existing)
            else:
                ev.display_num = self._next_num
                self._id_map[str(self._next_num)] = ev.id
                self._next_num += 1
        self._evict_old_entries()
        self._save_id_map()

    # ------------------------------------------------------------------
    # Contacts
    # ------------------------------------------------------------------

    def get_contacts(self, top: int = 50) -> list[Contact]:
        resp = self._get("/contacts", params={"$top": top})
        return [Contact.from_api(c) for c in resp.get("value", [])]

    # ------------------------------------------------------------------
    # Categories
    # ------------------------------------------------------------------

    def get_master_categories(self) -> list[dict]:
        """Fetch master category list via OWA service.svc.

        REST v2 doesn't expose /outlook/masterCategories.
        OWA uses service.svc with the action FindCategoryDetails,
        sending the JSON payload URL-encoded in the x-owa-urlpostdata header.
        """
        return self._owa_action("FindCategoryDetails", {
            "__type": "FindCategoryDetailsJsonRequest:#Exchange",
            "Header": {
                "__type": "JsonRequestHeaders:#Exchange",
                "RequestServerVersion": "V2018_01_08",
                "TimeZoneContext": {
                    "__type": "TimeZoneContext:#Exchange",
                    "TimeZoneDefinition": {
                        "__type": "TimeZoneDefinitionType:#Exchange",
                        "Id": "UTC",
                    },
                },
            },
            "Body": {
                "__type": "FindCategoryDetailsRequest:#Exchange",
            },
        })

    def get_categories(self, message_id: str) -> list[str]:
        real_id = self._resolve_id(message_id)
        resp = self._get(f"/messages/{real_id}", params={"$select": "Categories"})
        return resp.get("Categories", [])

    def set_categories(self, message_id: str, categories: list[str]) -> list[str]:
        real_id = self._resolve_id(message_id)
        resp = self._patch(f"/messages/{real_id}", json={"Categories": categories})
        return resp.get("Categories", categories)

    def add_category(self, message_id: str, category: str) -> list[str]:
        current = self.get_categories(message_id)
        if category not in current:
            current.append(category)
        return self.set_categories(message_id, current)

    def remove_category(self, message_id: str, category: str) -> list[str]:
        current = self.get_categories(message_id)
        current = [c for c in current if c != category]
        return self.set_categories(message_id, current)

    # ------------------------------------------------------------------
    # User info
    # ------------------------------------------------------------------

    def get_me(self) -> dict:
        return self._get("")

    # ------------------------------------------------------------------
    # ID mapping
    # ------------------------------------------------------------------

    def _resolve_id(self, display_id: str) -> str:
        """Convert display number to real Outlook ID."""
        if display_id in self._id_map:
            return self._id_map[display_id]
        # Maybe it's already a real ID (long base64 string)
        if len(display_id) > 50:
            return display_id
        raise ResourceNotFoundError(
            f"Unknown message #{display_id}. Run 'outlook inbox' first to populate the ID map."
        )

    def _assign_display_nums(self, messages: list[Email]) -> None:
        for msg in messages:
            # Check if this real ID already has a display number
            existing = next(
                (k for k, v in self._id_map.items() if v == msg.id and k.isdigit()),
                None,
            )
            if existing:
                msg.display_num = int(existing)
            else:
                msg.display_num = self._next_num
                self._id_map[str(self._next_num)] = msg.id
                self._next_num += 1
        self._evict_old_entries()
        self._save_id_map()

    def _evict_old_entries(self) -> None:
        """Keep only the most recent MAX_ID_MAP_SIZE entries."""
        numeric = sorted(
            ((int(k), k) for k in self._id_map if k.isdigit()),
            key=lambda x: x[0],
        )
        if len(numeric) <= self.MAX_ID_MAP_SIZE:
            return
        to_remove = numeric[: len(numeric) - self.MAX_ID_MAP_SIZE]
        for _, k in to_remove:
            del self._id_map[k]

    def _load_id_map(self) -> dict[str, str]:
        if ID_MAP_FILE.exists():
            try:
                return json.loads(ID_MAP_FILE.read_text())
            except (json.JSONDecodeError, OSError):
                pass
        return {}

    def _save_id_map(self) -> None:
        CACHE_DIR.mkdir(parents=True, exist_ok=True)
        ID_MAP_FILE.write_text(json.dumps(self._id_map))

    # ------------------------------------------------------------------
    # HTTP helpers
    # ------------------------------------------------------------------

    def _get(self, path: str, params: dict | None = None) -> dict:
        return self._request("GET", path, params=params)

    def _post(self, path: str, json: dict | None = None) -> dict:
        return self._request("POST", path, json=json)

    def _patch(self, path: str, json: dict | None = None) -> dict:
        return self._request("PATCH", path, json=json)

    def _delete(self, path: str) -> dict:
        return self._request("DELETE", path)

    def _request(
        self,
        method: str,
        path: str,
        params: dict | None = None,
        json: dict | None = None,
        _retry: int = 0,
    ) -> dict:
        resp = self._client.request(method, path, params=params, json=json)

        if resp.status_code == 401:
            raise TokenExpiredError("Token expired. Run: outlook login")

        if resp.status_code == 429:
            if _retry >= 3:
                raise RateLimitError("Rate limited after 3 retries")
            retry_after = int(resp.headers.get("Retry-After", 2 ** (_retry + 1)))
            time.sleep(retry_after)
            return self._request(method, path, params=params, json=json, _retry=_retry + 1)

        if resp.status_code == 204:
            return {}

        resp.raise_for_status()

        if not resp.content:
            return {}
        return resp.json()


    def _owa_action(self, action: str, payload: dict) -> dict:
        """Call OWA service.svc endpoint.

        OWA uses a non-standard pattern: the JSON payload is URL-encoded
        in the x-owa-urlpostdata header, and the body is empty.
        """
        resp = httpx.post(
            f"{OWA_SERVICE_URL}?action={action}",
            headers={
                "Authorization": f"Bearer {self._token}",
                "User-Agent": USER_AGENT,
                "Content-Type": "application/json; charset=utf-8",
                "Action": action,
                "x-req-source": "Mail",
                "x-owa-urlpostdata": quote(json.dumps(payload), safe=""),
            },
            content=b"",
            timeout=15,
        )
        if resp.status_code == 401:
            raise TokenExpiredError("Token expired. Run: outlook login")
        resp.raise_for_status()
        return resp.json()


